"""
# Tense (`AveyTense` on PyPi)

\\@since 0.3.24 \\
© 2023-Present Aveyzan // License: MIT
```ts
module tense
```
Multipurpose library with several extensions, including for built-ins.

Documentation: https://aveyzan.glitch.me/tense/py.html

Submodules:
- `tense.types_collection` - types collection
- `tense.constants` - constants collection
- `tense.fencord` - connection with discord
- `tense.operators` (since 0.3.27a3) - extension of `operator` library
- `tense.databases` (since 0.3.27a4) - connection with SQL. *Experimental*
"""

import sys as _sys

if _sys.version_info < (3, 9):
    error = (RuntimeError, "To use TensePy library, consider having Python 3.9 or newer")
    raise error

# 0.3.34: Prevent internal imports and prevent these imported subsequently
import ast as _ast
import bisect as _bisect
import builtins as _builtins
import collections as _collections
import copy as _copy
import dis as _dis
import inspect as _inspect
import io as _io
import itertools as _itertools
import math as _math
import os as _os
import pickle as _pickle
import platform as _platform
import random as _random
import re as _re
import secrets as _secrets
import socket as _socket
import string as _string
import time as _time
import timeit as _timeit
import tkinter as _tkinter # (< 0.3.36)
import types as _types
import typing as _typing
import uuid as _uuid
import warnings as _warnings

from . import _abroad as _ab_mod
from . import util as _util
from . import types_collection as _tc
from ._primal import *
from .types_collection import Any as _Any

__module__ = "tense"

# between @since and @author there is unnecessarily long line spacing
# hence this warning is being thrown; it is being disabled.
_warnings.filterwarnings("ignore")

_E = {
    "string_op_right": "expected a string as a right operand",
    "ls_error": "expected a list or a string",
    "to_list_error": "expected an iterable object or instance of '{}' or '{}'".format(_tc.ListConvertible.__name__, _tc.TupleConvertible.__name__)
}

_var = _tc.TypeVar
_par = _tc.ParamSpec
_uni = _tc.Union
_lit = _tc.Literal
_opt = _tc.Optional
_cal = _tc.Callable

_S = _var("_S")
_T = _var("_T")
_T1 = _var("_T1")
_T2 = _var("_T2")
_T3 = _var("_T3")
_KT = _var("_KT")
_KT1 = _var("_KT1")
_KT2 = _var("_KT2")
_VT = _var("_VT")
_VT1 = _var("_VT1")
_VT2 = _var("_VT2")
_V1 = _var("_V1")
_V2 = _var("_V2")
_M = _var("_M")
_P = _par("_P")
_Ts = _tc.TypeVarTuple("_Ts")
_T_richComparable = _var("_T_richComparable", bound = _tc.RichComparable)
_T_func = _var("_T_func", bound = _tc.Callable[..., _tc.Any])


_Bits = _lit[3, 4, 8, 24]
_ProbabilityType = _tc.ProbabilityType[_T] # expected int
_ProbabilitySeqNoDict = _uni[list, _collections.deque, set, frozenset, tuple]
_ProbabilitySeq = _uni[_ProbabilitySeqNoDict, dict]
_SizeableItemGetter = _tc.PickSequence[_T]
_FileType = _tc.FileType
_FileMode = _tc.FileMode
_FileOpener = _tc.FileOpener
_HaveCodeType = _uni[
    # satisfy dis.dis() first argument requirement
    str, bytes, bytearray, _types.MethodType, _types.FunctionType, _tc.CodeType, type, _tc.AnyCallable, None
]

_Statement = _tc.Union[_tc.Callable[[], object], str]
_Timer = _tc.Callable[[], float]
_GroupMode = _lit["and-or", "or-and", "and-nor", "nor-and", "nand-or", "or-nand", "nand-nor", "nor-nand", "and-and", "or-or", "nand-nand", "nor-nor"] # 0.3.34

_ReckonNGT = _ab_mod.ReckonNGT
_AbroadValue1 = _ab_mod.AbroadValue1[_T]
_AbroadValue2 = _ab_mod.AbroadValue2[_T]
_AbroadModifier = _ab_mod.AbroadModifier[_T]
_AbroadPackType = _ab_mod.AbroadPackType[_T]
_AbroadConvectType = _ab_mod.AbroadConvectType[_T]
_AbroadLiveType = _ab_mod.AbroadLiveType[_T]
_AbroadVividType = _ab_mod.AbroadVividType[_V1, _V2, _M]
_AbroadStringInitializer = _ab_mod.AbroadStringInitializer
_AbroadFloatyInitializer = _ab_mod.AbroadFloatyInitializer
_AbroadMultiInitializer = list[list[int]]
_AbroadImmutableInitializer = tuple[int, ...]
_AbroadIntFloatInitializer = list[_T]

NULL = _tc.NULL
"""\\@since 0.3.34; from `tense.types_collection`"""
# class _AbroadHexMode(tc.IntegerFlag): ### to 0.3.27
class _AbroadHexMode(_tc.Enum):
    "\\@since 0.3.26rc2. Internal class for class method `Tense.abroadHex()` parameter `mode`"
    INCLUDE = 0
    INCLUDE_HASH = 1
    EXCLUDE = 2

if False: # can be shortened to _AbroadHexMode (0.3.27)
    _AbroadHexModeType = _lit[
        _AbroadHexMode.INCLUDE,
        _AbroadHexMode.INCLUDE_HASH,
        _AbroadHexMode.EXCLUDE
    ]
    "\\@since 0.3.26rc2"

_AbroadEachType = _opt[_cal[[int], _T]]
"\\@since 0.3.26rc2"

_builtins_type = type

def _architecture(executable = _sys.executable, bits = "", linkage = ""):
    "\\@since 0.3.26rc2"
    
    return _platform.architecture(executable, bits, linkage)[0]

def _is_hexadecimal(target: str, /):
    _t = target
    HEX = _string.hexdigits
    
    if target.startswith(("0x", "#")):
        _t = _re.sub(r"^(0x|#)", "", _t)
    
    for c in target:
        
        if c not in HEX:
            return False
        
    return True

def _is_decimal(target: str, /):
    DEC = _string.digits
    
    for c in target:
        
        if c not in DEC:
            return False
        
    return True

def _is_octal(target: str, /):
    _t = target
    OCT = "01234567"
    
    if target.startswith("0o"):
        _t = _re.sub(r"^0o", "", _t)
    
    for c in target:
        
        if c not in OCT:
            return False
        
    return True

def _is_binary(target: str, /):
    _t = target
    BIN = "01"
    
    if target.startswith("0b"):
        _t = _re.sub(r"^0b", "", _t)
    
    for c in target:
        
        if c not in BIN:
            return False
        
    return True

def _is_condition_callback(v: object, /) -> _tc.TypeIs[_cal[[_Any], bool]]:
    return _inspect.isfunction(v)

def _int_conversion(target: str, /):
        
    if _is_hexadecimal(target):
        return int(target, 16)
    
    elif _is_decimal(target):
        return int(target, 10)
    
    elif _is_octal(target):
        return int(target, 8)
    
    elif _is_binary(target):
        return int(target, 2)
    
    else:
        return int(target)

def _get_all_params(f: _T_func, /):
    
    if hasattr(f, "__annotations__"):
        return [k for k in f.__annotations__ if k not in ("self", "return")]
    
    else:
        return [k for k in _inspect.get_annotations(f) if k not in ("self", "return")]


class _ProbabilityLength(_tc.Enum):
    "\\@since 0.3.26rc2. Internal class for class `tense.Tense`, for probability methods"
    
    PROBABILITY_COMPUTE = -1
    PROBABILITY_DEFAULT = 10000
    
    if _architecture() == "64bit":
        PROBABILITY_MAX = 2 ** 63 - 1 # 9223372036854775807
        
    else:
        PROBABILITY_MAX = 2 ** 31 - 1 # 2147483647
    
    PROBABILITY_MIN = 1

class _BisectMode(_tc.Enum):
    "\\@since 0.3.26rc2. Internal class for method `Tense.bisect()`"
    BISECT_LEFT = 0
    BISECT_RIGHT = 1

class _InsortMode(_tc.Enum):
    "\\@since 0.3.26rc2. Internal class for method `Tense.insort()`"
    INSORT_LEFT = 0
    INSORT_RIGHT = 1
    
_ProbabilityLengthType = _uni[int, _lit[_ProbabilityLength.PROBABILITY_COMPUTE]]
            
if False:
    class ParamVar:
        
        def __new__(cls, f: _cal[_P, _T], /):
            """
            \\@since 0.3.? (in code since 0.3.33)
            
            Returned dictionary with following keywords:
            - `p_k` - positional or keyword arguments
            - `p_only` - positional only arguments
            - `k_only` - keyword only arguments
            - `p_var` - variable positional argument (preceded with `*`)
            - `k_var` - variable keyword argument (preceded with `**`)
            """
            
            # temporary to deduce type Any
            _a = []
            _a.append("")
            
            _internal = {
                "tmp": [("", ("", _a[0]), _a[0])]
            }
            
            del _a, _internal["tmp"]
            
            _LACK_DEFAULT = "<lack>"
            _sig = _inspect.signature(f)
            
            _internal["p_k"] = [
                (_sig.parameters[k].name,
                (str(type(_sig.parameters[k].annotation)), _sig.parameters[k].annotation),
                _sig.parameters[k].default if _sig.parameters[k].default is _sig.parameters[k].empty else _LACK_DEFAULT) for k in _sig.parameters if _sig.parameters[k].kind == _sig.parameters[k].POSITIONAL_OR_KEYWORD
            ]
            
            _internal["p_only"] = [
                (_sig.parameters[k].name,
                (str(type(_sig.parameters[k].annotation)), _sig.parameters[k].annotation),
                _sig.parameters[k].default if _sig.parameters[k].default is _sig.parameters[k].empty else _LACK_DEFAULT) for k in _sig.parameters if _sig.parameters[k].kind == _sig.parameters[k].POSITIONAL_ONLY
            ]
            
            _internal["k_only"] = [
                (_sig.parameters[k].name,
                (str(type(_sig.parameters[k].annotation)), _sig.parameters[k].annotation),
                _sig.parameters[k].default if _sig.parameters[k].default is _sig.parameters[k].empty else _LACK_DEFAULT) for k in _sig.parameters if _sig.parameters[k].kind == _sig.parameters[k].KEYWORD_ONLY
            ]
            
            _internal["p_var"] = [
                (_sig.parameters[k].name,
                (str(type(_sig.parameters[k].annotation)), _sig.parameters[k].annotation),
                _LACK_DEFAULT) for k in _sig.parameters if _sig.parameters[k].kind == _sig.parameters[k].VAR_POSITIONAL
            ]
            
            _internal["k_var"] = [
                (_sig.parameters[k].name,
                (str(type(_sig.parameters[k].annotation)), _sig.parameters[k].annotation),
                _LACK_DEFAULT) for k in _sig.parameters if _sig.parameters[k].kind == _sig.parameters[k].VAR_KEYWORD
            ]
            
            return _internal


class TenseOptions(_util.AbstractFinal):
    """
    \\@since 0.3.27a5
    ```ts
    in module tense
    ```
    Several settings holder class. Cannot be initialized nor subclassed.
    """
    initializationMessage = False
    """
    \\@since 0.3.27a5
    
    Toggle on/off initialization message in the terminal
    """
    
    insertionMessage = False
    """
    \\@since 0.3.27b1
    
    Toggle on/off insertion messages (these displayed by `Tense.print()`). \\
    If this option was `False`, invoked is mere `print()` method. This option \\
    has also influence on `Fencord` solutions.
    """
    if False: # to be toggled on on version 0.3.27 or later
        probabilityExtendedLength = False
        """
        \\@since 0.3.27rc1. Toggle on/off extended length via 2d list technique. Once it is toggled off, \\
        error will be thrown if going above `sys.maxsize`, `(sys.maxsize + 1) ** 2 - 1` otherwise.
        """
    
    disableProbability2LengthLimit = False
    """
    \\@since 0.3.31
    
    Switch on/off length limit, which bases on `sys.maxsize`. If that option was enabled, length passed \\
    to `Tense.probability2()` class method will no longer throw an error whether length is greater than \\
    `sys.maxsize`. This also prevents creating a temporary list to generate the result via this method. It will \\
    not work on extended variation of this method, `Tense.probability()`, unless there were exactly 2 \\
    integer values passed. Its default value is `False`
    """
    
    enableFurryDeclarations = False
    """
    \\@since 0.3.34
    
    When this setting is `True`, normally `Tense.owoify()` and `Tense.uwuify()` will be available at runtime. \\
    Using these, when this option is `False`, throw an error. Default value is `False`
    
    """


class FencordOptions(_util.AbstractFinal):
    """
    \\@since 0.3.27b1
    ```py
    from tense import FencordOptions
    ```
    Several settings holder class. Cannot be initialized nor subclassed.
    """
    
    initializationMessage = False
    "\\@since 0.3.27b1. Toggle on/off initialization message in the terminal"


class Tense(
    
    # NennaiAbroads, ### < 0.3.34
    # NennaiStringz, ### no inherit since 0.3.27
    # NennaiRandomize, ### < 0.3.34
    Time,
    Math,
    # _tc.Positive[str], ### < 0.3.34
    # _tc.Negative[str], ### < 0.3.34
    # _tc.Invertible[str], ### < 0.3.34
    _tc.BitwiseLeftOperable,
    _tc.BitwiseRightOperable
    
):
    """
    \\@since 0.3.24 (standard since 0.3.24)
    ```py
    class Tense(Math, Time): ...
    from tense import Tense
    ```
    Root of TensePy. Subclassing since 0.3.26b3
    """
    from .constants import (
        VERSION as version,
        VERSION_ID as versionId,
        VERSION_TUPLE as versionTuple
    ) # last 2 = since 0.3.27a2
    
    PROBABILITY_COMPUTE = _ProbabilityLength.PROBABILITY_COMPUTE
    "\\@since 0.3.26rc2"
    
    BISECT_LEFT = _BisectMode.BISECT_LEFT
    "\\@since 0.3.26rc2"
    
    BISECT_RIGHT = _BisectMode.BISECT_RIGHT
    "\\@since 0.3.26rc2"
    
    INSORT_LEFT = _InsortMode.INSORT_LEFT
    "\\@since 0.3.26rc2"
    
    INSORT_RIGHT = _InsortMode.INSORT_RIGHT
    "\\@since 0.3.26rc2"
    
    TIMEIT_TIMER_DEFAULT = _time.perf_counter
    "\\@since 0.3.26rc3"
    
    TIMEIT_NUMBER_DEFAULT = 1000000
    "\\@since 0.3.26rc3"
    
    streamLikeC = False
    "\\@since 0.3.27a5. If set to `True`, output becomes `<<` and input - `>>`, vice versa otherwise"
    
    streamInputPrompt = ""
    "\\@since 0.3.27a5. Prompt for `input()` via `>>` and `<<` operators, depending on value of setting `streamLikeC`"
    
    streamInputResult = ""
    "\\@since 0.3.27b1. Result from `>>` or `<<`, depending on which one of them is for input"
    
    # __formername__ = "Tense08" ### < 0.3.34
    
    def __init__(self):
        e = super().fencordFormat()
        
        if TenseOptions.initializationMessage:
            print(f"\33[1;90m{e}\33[1;36m INITIALIZATION\33[0m Class '{__class__.__name__}' was successfully initalized. Line {_inspect.currentframe().f_back.f_lineno}")
            
    if False: # since 0.3.34
        def __str__(self):
            e = super().fencordFormat()
            
            if self.__formername__ != "Tense08":
                error = ValueError(f"when invoking string constructor of '{__class__.__name__}', do not rename variable '__formername__'")
                raise error
            
            try:
                subcl = f"'{__class__.__subclasses__()[0]}', "
                for i in abroad(1, __class__.__subclasses__()):
                    subcl += f"'{__class__.__subclasses__()[i]}', "
                subcl = _re.sub(r", $", "", subcl)
                
            except IndexError(AttributeError):
                subcl = f"'{NennaiAbroads.__name__}', '{NennaiRandomize.__name__}', '{Math.__name__}', '{Time.__name__}'"
                
            return f"""
                \33[1;90m{e}\33[1;38;5;51m INFORMATION\33[0m Basic '{__class__.__name__}' class information (in module 'tense')

                Created by Aveyzan for version 0.3.24 as a deputy of cancelled class '{__class__.__formername__}'. The '{__class__.__name__}' class is a subclass of various classes located inside other
                Tense files: {subcl}. Class itself wasn't able to be subclassed (until 0.3.26b3). Generally speaking, the '{__class__.__name__}' class is a collection of many various methods inherited
                from all of these classes, but also has some defined within its body itself, like methods: probability(), random(), pick() etc.
            """
    
        def __pos__(self):
            "Return information about this class. Since 0.3.26rc1"
            return self.__str__()
        
        def __neg__(self):
            "Return information about this class. Since 0.3.26rc1"
            return self.__str__()
        
        def __invert__(self):
            "Return information about this class. Since 0.3.26rc1"
            return self.__str__()
    
    def __lshift__(self, other: object):
        "\\@since 0.3.27a5. A C-like I/O printing"
        
        if self.streamLikeC:
            self.print(other)
            
        else:
            if self.isString(other):
                self.streamInputResult = input(self.streamInputPrompt)
            else:
                error = TypeError(_E["string_op_right"])
                raise error
            
        return self
    
    def __rshift__(self, other: object):
        "\\@since 0.3.27a5. A C-like I/O printing"
        
        if not self.streamLikeC:
            self.print(other)
            
        else:
            if self.isString(other):
                self.streamInputResult = input(self.streamInputPrompt)
            else:
                error = TypeError(_E["string_op_right"])
                raise error
            
        return self
    
    @property
    def none(self):
        """
        \\@since 0.3.32
        
        This property is console-specific, and simply returns `None`.
        """
        return None
    
    ABROAD_HEX_INCLUDE = _AbroadHexMode.INCLUDE
    "\\@since 0.3.26rc2"
    
    ABROAD_HEX_INCLUDE_HASH = _AbroadHexMode.INCLUDE_HASH
    "\\@since 0.3.26rc2"
    
    ABROAD_HEX_EXCLUDE = _AbroadHexMode.EXCLUDE
    "\\@since 0.3.26rc2"
    
    @classmethod
    def abroadPositive(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None):
        """
        \\@since 0.3.24 \\
        \\@modified 0.3.25 (moved slash to between `value1` and `value2`), 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Every negative integer is coerced to positive.
        """
        ab = abroad(value1, value2, modifier)
        return type(ab)([abs(e) for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadNegative(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None):
        """
        \\@since 0.3.24 \\
        \\@modified 0.3.25 (moved slash to between `value1` and `value2`), 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Every positive integer is coerced to negative.
        """
        ab = abroad(value1, value2, modifier)
        return type(ab)([-abs(e) for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadPositiveFlip(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None):
        """
        \\@since 0.3.24 \\
        \\@modified 0.3.25 (moved slash to between `value1` and `value2`), 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Every negative integer is coerced to positive, then sequence is reversed.
        """
        ab = abroad(value1, value2, modifier)
        return type(ab)([abs(e) for e in ab][::-1], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadNegativeFlip(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None):
        """
        \\@since 0.3.24 \\
        \\@modified 0.3.25 (moved slash to between `value1` and `value2`), 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Every positive integer is coerced to negative, then sequence is reversed.
        """
        ab = abroad(value1, value2, modifier)
        return type(ab)([-abs(e) for e in ab][::-1], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadPack(self, *values: _AbroadPackType[_T]):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function bases on `zip()` Python function.
        """
        ab = abroad(reckonLeast(*values))
        return type(ab)([e for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadExclude(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, *excludedIntegers: int):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function is the same as `abroad()` function, \\
        but it also allows to exclude specific integers from the returned list. \\
        If all are excluded, returned is empty integer list. If integers excluded \\
        do not exist in returned sequence normally, this issue is omitted.
        """
        for e in excludedIntegers:
            if not isinstance(e, int):
                error = TypeError("every item in parameter 'excludedIntegers' must be an integer")
                raise error
        
        ab = abroad(value1, value2, modifier)
        return type(ab)([e for e in ab if e not in excludedIntegers], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadPrecede(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, prefix: _opt[str] = None):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns strings in a list. If `prefix` is `None`, \\
        returned are integers in strings, otherwise added is special string prefix before integers.
        """
        if prefix is not None and not isinstance(prefix, str):
            error = TypeError("expected parameter '{}' have string value".format(_get_all_params(self.abroadPrecede)[-1]))
            raise error

        ab = abroad(value1, value2, modifier)
        return _AbroadStringInitializer([str(e) for e in ab] if prefix is None else [prefix + str(e) for e in ab], ab.params[0], ab.params[1], ab.params[2])
            
    @classmethod
    def abroadSufcede(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, suffix: _opt[str] = None):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns strings in a list. If `prefix` is `None`, \\
        returned are integers in strings, otherwise added is special string suffix after integers.
        """
        if suffix is not None and not isinstance(suffix, str):
            error = TypeError("expected parameter '{}' have string value".format(_get_all_params(self.abroadSufcede)[-1]))
            raise error

        ab = abroad(value1, value2, modifier)
        return _AbroadStringInitializer([str(e) for e in ab] if suffix is None else [str(e) + suffix for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadInside(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, string: _opt[str] = None):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns strings in a list. If `string` is `None`, \\
        returned are integers in strings, otherwise integers are placed inside `{}` of the string.
        """
        if string is not None and not isinstance(string, str):
            error = TypeError("expected parameter '{}' have string value".format(_get_all_params(self.abroadInside)[-1]))
            raise error

        ab = abroad(value1, value2, modifier)
        return _AbroadStringInitializer([str(e) for e in ab] if string is None else [string.format(str(e)) for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    if False: # removed 0.3.34 due to use of -abroad()
        @classmethod
        @_tc.deprecated("Pending removal on 0.3.30 due to reorganization of sequences returned by abroad() function and many of its variations (retrieve tuple via -abroad(...)) during 0.3.28 - 0.3.29")
        def abroadImmutable(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None) -> _AbroadImmutableInitializer:
            """
            \\@since 0.3.25
            ```ts
            // created 09.07.2024
            // to 0.3.34 in class NennaiAbroads
            "class method" in class Tense
            ```
            Immutable variation of `abroad()` function - instead of list returned is tuple. \\
            Equals `tuple(abroad(...))`.
            """
            return tuple(abroad(value1, value2, modifier))
    
    @classmethod
    def abroadConvect(self, *values: _AbroadConvectType[_T]):
        """
        \\@since 0.3.25
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Typical math sum operation before returned is list from `abroad()` function. \\
        If from values a value is:
        - an integer - added is this integer
        - a float - added is this number, without fraction
        - a complex - added are both real and imaginary parts
        - sizeable object - added is its length

        Notice: you can also provide negative entities! If resulted number is negative, \\
        up to `abroad()` function, sequence will go in range `[values_sum, -1]`. \\
        Otherwise, it will take this form: `[0, values_sum - 1]`.
        """
        i = 0
        _params = _get_all_params(self.abroadConvect)
        
        if reckon(values) == 0:
            error = _tc.MissingValueError("expected at least one item in parameter '{}'".format(_params[-1]))
            raise error
        
        for e in values:
            
            if not isinstance(e, (_ReckonNGT, int, float, complex)):
                error = TypeError("from gamut of supported types, parameter '{}' has at least one unsupported type".format(_params[-1]))
                raise error
            
            elif isinstance(e, int):
                i += e
                
            elif isinstance(e, float):
                i += _math.trunc(e)
                
            elif isinstance(e, complex):
                i += _math.trunc(e.real) + _math.trunc(e.imag)
                
            else:
                i += reckon(e)
        return abroad(i)
    
    @classmethod
    def abroadLive(self, *values: _AbroadLiveType[_T]):
        """
        \\@since 0.3.25
        ```ts
        // created 09.07.2024ts
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Concept from non-monotonous sequences from math. Like graph, \\
        which changes per time. If from values a value is:
        - an integer - this is next point
        - a float - next point doesn't have fraction
        - a complex - next point is sum of real and imaginary parts
        - sizeable object - its length is next point
        """
        a, ret = [[0] for _ in abroad(2)]
        for e in (a, ret): e.clear()
        
        if reckon(values) == 0:
            error = _tc.MissingValueError("expected at least one item in parameter 'values'.")
            raise error
        
        for e in values:
            
            if not isinstance(e, (_ReckonNGT, int, float, complex)):
                error = TypeError(f"from gamut of supported types, parameter 'values' has at least one unsupported type: '{type(e).__name__}'")
                raise error
            
            elif isinstance(e, int):
                a.append(e)
                
            elif isinstance(e, float):
                a.append(_math.trunc(e))
                
            elif isinstance(e, complex):
                a.append(_math.trunc(e.real) + _math.trunc(e.imag))
            else:
                a.append(reckon(e))
                
        for i1 in abroad(1, a):
            tmp = a[i1]
            if tmp < 0: tmp -= 1
            else: tmp += 1
            
            for i2 in abroad(a[i1 - 1], tmp): 
                ret.append(i2)
                
        return ret
    
    @classmethod
    def abroadFloaty(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, div: _tc.FloatOrInteger = 10):
        """
        \\@since 0.3.25 \\
        \\@modified 0.3.29
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Every item from `abroad()` function will be divided by parameter `div`. \\
        It's default value is `10`.
        """
        if not isinstance(div, (int, float)):
            error = TypeError(f"parameter 'div' is not an integer nor floating-point number. Ensure argument got value of type 'int' or 'float'. Received type: {type(div).__name__}")
            raise error
        
        elif isinstance(div, float) and div in (_math.nan, _math.inf):
            error = ValueError("parameter 'div' may not be infinity or not a number.")
            raise error
        
        elif (isinstance(div, int) and div == 0) or (isinstance(div, float) and div == .0):
            error = ZeroDivisionError("parameter 'div' may not be equal zero. This is attempt to divide by zero")
            raise error
        
        ab = abroad(value1, value2, modifier)
        return _AbroadFloatyInitializer([e / div for e in ab], ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadSplit(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, limit = 2) -> _AbroadMultiInitializer:
        """
        \\@since 0.3.25
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Reference to string slicing. Limit is amount of items, \\
        which can be in one sub-list. May not be equal or below 1.
        """
        lim = 0
        tmp, a = ([0], [[0]])
        tmp.clear()
        a.clear()
        
        if not isinstance(limit, int):
            error = TypeError(f"parameter 'limit' is not an integer. Ensure argument got integer value. Received type: {type(limit).__name__}")
            raise error
        
        elif limit < 1:
            error = ValueError("parameter 'limit' may not be negative, or have value 0 or 1. Start from 2.")
            raise error
        
        for i in abroad(value1, value2, modifier):
            if lim % limit == 0:
                a.append(tmp)
                tmp.clear()
            else:
                tmp.append(i)
            lim += 1
        return a
    
    @classmethod
    def abroadVivid(self, *values: _AbroadVividType[_V1, _V2, _M]) -> _AbroadMultiInitializer:
        """
        \\@since 0.3.25
        ```ts
        // created 09.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        For every value in `values` returned is list `[abroad(V1_1, V2_1?, M_1?), abroad(V1_2, V2_2?, M_2?), ...]`. \\
        Question marks are here to indicate optional values.
        """
        a = [[0]]
        a.clear()
        
        if reckon(values) < 2:
            error = ValueError("expected at least 2 items in parameter 'values'.")
            raise error
        
        for e in values:
            
            if not isinstance(e, tuple):
                error = TypeError(f"parameter 'values' has an item, which isn't a tuple. Ensure every item is a tuple. Received type: {type(e).__name__}")
                raise error
            
            if reckon(e) == 1:
                a.append(+abroad(e[0]))
                
            elif reckon(e) == 2:
                a.append(+abroad(e[0], e[1]))
                
            elif reckon(e) == 3:
                a.append(+abroad(e[0], e[1], e[2]))
                
            else:
                error = ValueError("parameter 'values' may not have empty tuples, nor tuples of size above 3.")
                raise error
            
        return a
    
    @classmethod
    def abroadEach(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, each: _AbroadEachType[_T] = None) -> _AbroadIntFloatInitializer[_T]:
        """
        \\@since 0.3.25 (experimental for 0.3.25 - 0.3.26b1)
        ```ts
        / created 10.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        Invoked is `each` callback for every item in `abroad()` function.
        """
        
        if each is None:
            a = [0]
            
        else:
            a = [0] + [0.0]
            
        a.clear()
        
        for i in abroad(value1, value2, modifier):
            
            if each is None:
                a.append(i)
            
            else:
                tmp = each(i)
                if not isinstance(tmp, (int, float)):
                    error = ValueError("parameter 'each' returns invalid type or has invalid parameter type (expected 'int' or 'float'), has too much parameters, or is not a callable object. Use 'lambda' expression, like 'lambda x: Tense.cbrt(x)'.")
                    raise error
                
                a.append(tmp)
        return a
    
    @classmethod
    def abroadHex(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, mode: _AbroadHexMode = ABROAD_HEX_INCLUDE):
        """
        \\@since 0.3.25
        ```ts
        // created 10.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns hexadecimal representation of each integer.

        Modes (for 0.3.26rc2; to 0.3.27 support for integers):
        - `self.ABROAD_HEX_INCLUDE` - appends `0x` to each string. It faciliates casting to integer.
        - `self.ABROAD_HEX_INCLUDE_HASH` - appends `#` to each string. Reference from CSS.
        - `self.ABROAD_HEX_EXCLUDE` - nothing is appended.
        """
        a, ab = ([""], abroad(value1, value2, modifier))
        a.clear()
        
        for i in ab:
            
            if not isinstance(mode, _AbroadHexMode):
                error = ValueError("expected a constant preceded with 'ABROAD_HEX_'")
                raise error
            
            elif mode == self.ABROAD_HEX_INCLUDE:
                a.append(hex(i))
                
            elif mode == self.ABROAD_HEX_INCLUDE_HASH:
                a.append(_re.sub(r"^0x", "#", hex(i)))
                
            else:
                a.append(_re.sub(r"^0x", "", hex(i)))
        
        return _AbroadStringInitializer(a, ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadBinary(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None, include_0b = True):
        """
        \\@since 0.3.25
        ```ts
        // created 10.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns binary representation of each integer. \\
        Parameter `include_0b` allows to append `0b` before binary notation, what allows \\
        to faciliate casting to integer. Defaults to `True`
        """
        a, ab = ([""], abroad(value1, value2, modifier))
        a.clear()
        
        for i in ab:
            
            if not isinstance(include_0b, bool):
                error = TypeError("expected parameter 'include_0b' to be of type 'bool'.")
                raise error
            
            elif include_0b:
                a.append(bin(i))
                
            else:
                a.append(_re.sub(r"^0b", "", bin(i)))
                
        return _AbroadStringInitializer(a, ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def abroadOctal(self, value1: _AbroadValue1[_T1], /, value2: _AbroadValue2[_T2] = None, modifier: _AbroadModifier[_T3] = None):
        """
        \\@since 0.3.25
        ```ts
        // created 18.07.2024
        // to 0.3.34 in class NennaiAbroads
        "class method" in class Tense
        ```
        This variation of `abroad()` function returns octal representation of each integer. \\
        Every string will be preceded with `0o`
        """
        a, ab = ([""], abroad(value1, value2, modifier))
        a.clear()
        
        for i in ab:
            # if not isinstance(include_0o, bool):
            #    err, s = (TypeError, "Expected parameter 'include_0o' to be of type 'bool'.")
            #    raise err(s)
            # elif include_0o:
                a.append(oct(i))
            # else:
            #   a.append(re.sub(r"^0o", "", oct(i)))
        return _AbroadStringInitializer(a, ab.params[0], ab.params[1], ab.params[2])
    
    @classmethod
    def toList(self, v: _tc.Union[_tc.Iterable[_T], _tc.ListConvertible[_T], _tc.TupleConvertible[_T]], /):
        """
        \\@since 0.3.26rc3
        ```ts
        "class method" in class Tense
        ```
        Converts a value to a `list` built-in.
        """
        
        if isinstance(v, _tc.ListConvertible):
            return v.__tlist__()
        
        elif isinstance(v, _tc.TupleConvertible):
            return list(v.__ttuple__())
        
        elif isinstance(v, _tc.Iterable):
            return list(v)
        
        else:
            error = TypeError(_E["to_list_error"])
            raise error
    
    @classmethod
    def toString(self, v: _tc.Object = ..., /):
        """
        \\@since 0.3.26rc3
        ```ts
        "class method" in class Tense
        ```
        Alias to `Tense.toStr()`, `str()`

        Converts a value to a `str` built-in.
        """
        return str(v)
    
    @classmethod
    def toStr(self, v: _tc.Object = ..., /):
        """
        \\@since 0.3.26rc3
        ```ts
        "class method" in class Tense
        ```
        Converts a value to a `str` built-in.
        """
        return str(v)
    
    @classmethod
    def isNone(self, v, /) -> _tc.TypeIs[None]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is `None`
        """
        return v is None
    
    @classmethod
    def isEllipsis(self, v, /) -> _tc.TypeIs[_tc.Ellipsis]:
        """
        \\@since 0.3.26
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is `...`
        """
        return v is ...
    
    @classmethod
    def isBool(self, v: object, /) -> _tc.TypeIs[bool]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is of type `bool`
        """
        return type(v) is bool
    
    @classmethod
    def isBoolean(self, v: object, /) -> _tc.TypeIs[bool]:
        """
        \\@since 0.3.26rc1
        ```ts
        "class method" in class Tense
        ```
        Alias to `Tense.isBool()`

        Determine whether a value is of type `bool`
        """
        return type(v) is bool
    
    @classmethod
    def isInt(self, v: object, /) -> _tc.TypeIs[int]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is of type `int`
        """
        return type(v) is int
    
    @classmethod
    def isInteger(self, v: object, /) -> _tc.TypeIs[int]:
        """
        \\@since 0.3.26rc1
        ```ts
        "class method" in class Tense
        ```
        Alias to `Tense.isInt()`

        Determine whether a value is of type `int`
        """
        return type(v) is int
    
    @classmethod
    def isFloat(self, v: object, /) -> _tc.TypeIs[float]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is of type `float`
        """
        return type(v) is float
    
    @classmethod
    def isComplex(self, v: object, /) -> _tc.TypeIs[complex]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is of type `complex`
        """
        return type(v) is complex
    
    @classmethod
    def isStr(self, v: object, /) -> _tc.TypeIs[str]:
        """
        \\@since 0.3.26b3
        ```ts
        "class method" in class Tense
        ```
        Determine whether a value is of type `str`
        """
        return type(v) is str
    
    @classmethod
    def isString(self, v: object, /) -> _tc.TypeIs[str]:
        """
        \\@since 0.3.26rc1
        ```ts
        "class method" in class Tense
        ```
        Alias to `Tense.isStr()`

        Determine whether a value is of type `str`
        """
        return type(v) is str
    
    # CHANGEOVER 0.3.34
    if False:
        
        @classmethod
        @_tc.overload
        def isTuple(self, v: object, type: type[_T] = _Any, /) -> _tc.TypeIs[tuple[_T, ...]]:
            """
            \\@since 0.3.26rc1
            
            Determine whether a value is a tuple.
            
            - 0.3.34: Added new parameters `type` and `types`
            """
            ...
        
        # this overload is experimental to be finished
        @classmethod
        @_tc.overload
        @_tc.deprecated("The overload is experimental, use the other overload, without 'types' parameter (the overload is not actually deprecated)")
        def isTuple(self, v: object, type: type[_T], /, *types: _tc.Unpack[_Ts]) -> _tc.TypeIs[tuple[_T, _tc.Unpack[_Ts]]]:
            """
            \\@since 0.3.26rc1
            
            Determine whether a value is a tuple. Experimental due to unexpected `False` when passing more than \\
            one type.
            
            - 0.3.34: Added new parameters `type` and `types`
            """
            ...
    
        @classmethod
        def isTuple(self, v, type = _Any, /, *types):
            """
            \\@since 0.3.26rc1
            ```ts
            "class method" in class Tense
            ```
            Determine whether a value is of type `tuple`
            """
            
            if _builtins_type(v) is tuple:
                
                _placeholder = True
                
                if reckon(types) == 0:
                    
                    if reckon(v) == 0:
                        
                        if type is _Any:
                            return True
                        
                        return False
                    
                    if not isinstance(type, _builtins_type):
                        error = TypeError("passed value to parameter '{}' must be a type, not an object".format("type"))
                        raise error
                    
                    if type is _Any:
                        return True
                    
                    for e in v:
                        
                        try:
                            _placeholder = _placeholder and isinstance(e, type)
                            
                        except:
                            _placeholder = _placeholder and _builtins_type(e) is type
                
                    return _placeholder
                
                if reckon(v) != reckon(types) - 1: # minus one due to 'type' parameter
                    return False
                
                if type is _Any:
                    pass
                
                else:
                    
                    try:
                        _placeholder = _placeholder and isinstance(v[0], type)
                        
                    except:
                        _placeholder = _placeholder and _builtins_type(v[0]) is type
                
                _types_any_check = True
                
                for i in abroad(types):
                    _types_any_check = _types_any_check and types[i] is _Any
                    
                if _types_any_check and type is _Any:
                    return True
                
                for i in abroad(1, types):
                    
                    try:
                        _placeholder = _placeholder and isinstance(v[i], types[i - 1])
                        
                    except:
                        _placeholder = _placeholder and _builtins_type(v[i]) is types[i - 1]
                    
                return _placeholder
                
            else:
                return False
            
    else:
        
        @classmethod
        def isTuple(self, v: object, /, type: type[_T] = _Any) -> _tc.TypeIs[tuple[_T, ...]]:
            """
            \\@since 0.3.26rc1
            ```ts
            "class method" in class Tense
            ```
            Determine whether a value is a tuple built-in.
            
            - 0.3.34: Added new parameter `type`, allowing to restrict the tuple type. Default value is `Any`. \\
            WARNING: Update appending `types` parameter is currently during experiments and may be \\
            featured later. The `types` parameter has to restrict tuple more - its content basically.
            """
            if _builtins_type(v) is tuple:
                
                if reckon(v) == 0:
                    
                    if type is _Any:
                        return True
                    
                    return False
                
                if not isinstance(type, _builtins_type):
                    error = TypeError("passed value to parameter '{}' must be a type, not an object".format("type"))
                    raise error
                
                _placeholder = True
                
                for e in v:
                    
                    try:
                        _placeholder = _placeholder and isinstance(e, type)
                        
                    except:
                        _placeholder = _placeholder and _builtins_type(e) is type
                    
                return _placeholder
            
            else:
                return False
                        
        @classmethod
        def isList(self, v: object, /, type: type[_T] = _Any) -> _tc.TypeIs[list[_T]]:
            """
            \\@since 0.3.26rc1
            ```ts
            "class method" in class Tense
            ```
            Determine whether a value is a list built-in.
            
            - 0.3.34: Added new parameter `type`, allowing to restrict the list type. Default value is `Any`
            """
            if _builtins_type(v) is list:
                
                if reckon(v) == 0:
                    
                    if type is _Any:
                        return True
                    
                    return False
                
                if not isinstance(type, _builtins_type):
                    error = TypeError("passed value to parameter '{}' must be a type, not an object".format("type"))
                    raise error
                
                _placeholder = True
                
                for e in v:
                    
                    try:
                        _placeholder = _placeholder and isinstance(e, type)
                        
                    except:
                        _placeholder = _placeholder and _builtins_type(e) is type
                    
                return _placeholder
            
            else:
                return False
    
        @classmethod
        def isDict(self, v: object, /, ktype: type[_KT] = _Any, vtype: type[_VT] = _Any) -> _tc.TypeIs[dict[_KT, _VT]]:
            """
            \\@since 0.3.31
            ```ts
            "class method" in class Tense
            ```
            Determine whether a value is a dictionary built-in.
            
            - 0.3.34: Added 2 parameters `ktype` and `vtype`, restricting types for, respectively, keys and values. \\
              Both have default values `Any`.
            """
            # notice not all classes support isinstance() checking, hence this notation
            if type(v) is dict:
                
                if reckon(v) == 0:
                    
                    if ktype is _Any and vtype is _Any:
                        return True
                    
                    return False
                
                if ktype is _Any and vtype is _Any:
                    return True
                
                _params = _get_all_params(self.isDict)
                
                if not isinstance(ktype, type):
                    error = TypeError("passed value to parameter '{}' must be a type, not an object".format(_params[0]))
                    raise error
                
                if not isinstance(vtype, type):
                    error = TypeError("passed value to parameter '{}' must be a type, not an object".format(_params[1]))
                    raise error
                
                _placeholder = True
                # fact worth explaining: even if 'type[object]' is deduced, these are still on their type TypeVar counterparts
                _vk = (ktype, list(v.keys()))
                _vv = (vtype, list(v.values()))
                
                for _v in (_vk, _vv):
                    
                    for e in _v[1]:
                        
                        try:
                            # not all types can be inspected with isinstance()
                            _placeholder = _placeholder and isinstance(e, _v[0])
                            
                        except:
                            # the 'is' keyword usage is unnecessary, but some types tend to use '==' operator differently
                            _placeholder = _placeholder and type(e) is _v[0]
                        
                return _placeholder
                
            else:
                return False
            
            
    @classmethod
    @_tc.overload
    def hasattr(self, o: object, attr: str, /) -> bool: ...
    
    @classmethod
    @_tc.overload
    def hasattr(self, o: object, attr: tuple[str, ...], /, mode: _lit["and", "or"] = "or") -> bool: ...
            
    @classmethod
    def hasattr(self, o, attr, /, mode = "or"):
        """
        \\@since 0.3.34
        
        Returns `True` if object has specific attribute. Same as built-in function `hasattr()`, only 2 additional things are provided:
        - If `attr` is a string tuple, then it is deduced as `hasattr(o, attr[0]) or hasattr(o, attr[1]) or ...` \\
        Reference to how `isinstance()` works in case if types tuple is included instead of single type.
        - If `mode` is set to `"and"`, then it will occur as `hasattr(o, attr[0]) and hasattr(o, attr[1]) and ...` \\
        By default has value `"or"`. Note it works only for `attr` being a string tuple.
        """
        
        if self.isString(attr):
            
            return hasattr(o, attr)
        
        elif self.isTuple(attr, str):
            
            _r = True
            
            for many in attr:
                
                if mode == "or":
                    _r = _r or hasattr(o, many)
                    
                elif mode == "and":
                    _r = _r and hasattr(o, many)
                    
                else:
                    error = ValueError("unknown mode provided, expected \"and\" or \"or\"")
                    raise error
                
            return _r
        
        error = TypeError("expected a string tuple or a string in parameter 'attr'")
        raise error
                
            
    @classmethod
    def group(self, *statements: _uni[_tc.Sequence[bool], _tc.Uniqual[bool]], mode: _GroupMode = "and-or"):
        """
        \\@since 0.3.34
        
        Returns one boolean value combining all statements into one boolean value. \\
        Parameter `mode` determines about used logical operators inside and outside \\
        provided sequences. Possible values (with `and`, `or`, `nand` = `and not`, `nor` = `or not`):
        
        - `"and-or"` = `(a1 and a2 and ... and aN) or (b1 and b2 and ... and bN) or ...`
        - `"or-and"` = `(a1 or a2 or ... or aN) and (b1 or b2 or ... or bN) and ...`
        - `"and-nor"` = `not (a1 and a2 and ... and aN) or not (b1 and b2 and ... and bN) or not ...`
        - `"nor-and"` = `(not a1 or not a2 or not ... or not aN) and (not b1 or not b2 or not ... or not bN) and ...`
        - `"nand-or"` = `(not a1 and not a2 and not ... and not aN) or (b1 and not b2 and not ... and not bN) or ...`
        - `"or-nand"` = `not (a1 or a2 or ... or aN) and not (b1 or b2 or ... or bN) and not ...`
        - `"nand-nor"` = `not (not a1 and not a2 and not ... and not aN) or not (b1 and not b2 and not ... and not bN) or not ...`
        - `"nor-nand"` = `not (not a1 or not a2 or not ... or not aN) and not (not b1 or not b2 or not ... or not bN) and not ...`
        
        Note: using modes `"and-and"`, `"or-or"`, `"nand-nand"` and `"nor-nor"` is discouraged, \\
        but will be kept to save some time writing `and`, `or` and `not` operators
        """
        
        _modes = ("and-or", "or-and", "and-nor", "nor-and", "nand-or", "or-nand", "nand-nor", "nor-nand", "and-and", "or-or", "nand-nand", "nor-nor")
        
        if mode not in _modes:
            
            error = ValueError("expected a valid mode from following: {}".format(", ".join(_modes)))
            raise error
        
        for statement in statements:
            
            if not isinstance(statement, (_tc.Sequence, _tc.Uniqual)) or (isinstance(statement, (_tc.Sequence, _tc.Uniqual)) and not self.isList(list(statement), bool)):
                
                error = ValueError("expected non-empty sequence(s) with single boolean values, like list, tuple, set or frozenset")
                raise error
            
            
        _result = _subresult = True
        
        for statement in statements:
            
            if mode == "and-and":
                
                for s in statement:
                    _subresult = _subresult and s
                    
                _result = _result and _subresult
                
            elif mode == "and-nor":
                
                for s in statement:
                    _subresult = _subresult or not s
                    
                _result = _result and _subresult
                
            elif mode == "and-or":
                
                for s in statement:
                    _subresult = _subresult or s
                    
                _result = _result and _subresult
                
            elif mode == "nand-nand":
                
                for s in statement:
                    _subresult = _subresult and not s
                    
                _result = _result and not _subresult
                
            elif mode == "nand-nor":
                
                for s in statement:
                    _subresult = _subresult or not s
                    
                _result = _result and not _subresult
                
            elif mode == "nand-or":
                
                for s in statement:
                    _subresult = _subresult or s
                    
                _result = _result and not _subresult
                
            elif mode == "nor-and":
                
                for s in statement:
                    _subresult = _subresult and s
                    
                _result = _result or not _subresult
                
            elif mode == "nor-nand":
                
                for s in statement:
                    _subresult = _subresult and not s
                    
                _result = _result or not _subresult
                
            elif mode == "nor-nor":
                
                for s in statement:
                    _subresult = _subresult or not s
                    
                _result = _result or not _subresult
                
            elif mode == "or-and":
                
                for s in statement:
                    _subresult = _subresult and s
                    
                _result = _result or _subresult
                
            elif mode == "or-nand":
                
                for s in statement:
                    _subresult = _subresult and not s
                    
                _result = _result or _subresult
                
            else:
                
                for s in statement:
                    _subresult = _subresult or s
                    
                _result = _result or _subresult
                
            _subresult = True
            
        return _result
                
                
    
    @classmethod
    def architecture(self, executable = sys.executable, bits = "", linkage = ""):
        """
        \\@since 0.3.26rc2 (0.3.27a5: added optional parameters)
        ```
        "class method" in class Tense
        ```
        Returns system's architecture
        """
        return _architecture(executable, bits, linkage)
    
    @classmethod
    def disassemble(
        self,
        x: _HaveCodeType = None,
        /,
        file: _tc.Optional[_tc.IO[str]] = None,
        depth: _tc.Optional[int] = None,
        showCaches = False,
        adaptive = False,
        showOffsets = False
    ):
        """
        \\@since 0.3.26rc3
        ```
        "class method" in class Tense
        ```
        Detach code of a class, type, function, methods and other compiled objects. \\
        If argument `x` is `None` (by default is `None`), disassembled is last traceback. \\
        See [`dis.dis()`](https://docs.python.org/3/library/dis.html#dis.dis) \\
        Modified 0.3.31: added missing parameter `showOffsets`
        """
        _dis.dis(x, file = file, depth = depth, show_caches = showCaches, adaptive = adaptive, show_offsets = showOffsets)
        return self
    
    @classmethod
    def timeit(
        self,
        statement: _tc.Optional[_Statement] = None,
        /,
        setup: _tc.Optional[_Statement] = None,
        timer: _Timer = TIMEIT_TIMER_DEFAULT,
        number = TIMEIT_NUMBER_DEFAULT,
        globals: _tc.Optional[dict[str, _Any]] = None
        ):
        """
        \\@since 0.3.26rc3
        ```
        "class method" in class Tense
        ```
        See [`timeit.timeit()`](https://docs.python.org/3/library/timeit.html#timeit.timeit) \\
        Return time execution for specific code scrap (`statement`). Basic use::

            Tense.timeit(lambda: pow(3, 2)) # 0.06483080000180053
            Tense.timeit(lambda: math.pow(3, 2)) # 0.1697132999979658
            Tense.timeit(lambda: Tense.pow(3, 2)) # 0.26907890000074985
        """
        return _timeit.timeit(
            stmt = "pass" if self.isNone(statement) else statement,
            setup = "pass" if self.isNone(setup) else setup,
            timer = timer,
            number = number,
            globals = globals
        )
        
    @classmethod
    def socket(self, family: _uni[int, _socket.AddressFamily] = -1, type: _uni[int, _socket.SocketKind] = -1, proto = -1, fileno: _opt[int] = None):
        """
        \\@since 0.3.27a2
        ```
        "class method" in class Tense
        ```
        See [`socket.socket`](https://docs.python.org/3/library/socket.html#socket.socket)
        """
        return _socket.socket(family, type, proto, fileno)
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def shuffle(self, v: str) -> str: ...
    
    @classmethod
    @_tc.overload
    def shuffle(self, v: _ab_mod.AbroadInitializer) -> list[int]: ...
    
    @classmethod
    @_tc.overload
    def shuffle(self, v: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]]) -> list[_T]: ...
    
    @classmethod
    @_tc.overload
    def shuffle(self, v: _tc.Mapping[_KT, _VT]) -> dict[_KT, _VT]: ...
    
    @classmethod
    def shuffle(self, v):
        """
        \\@since 0.3.26rc1
        ```ts
        "class method" in class Tense
        ```
        Shuffle a string, mapping or a sequence, and return shuffled iterable \\
        without changing passed iterable.
        """
        
        if self.isString(v):
            _v = [c for c in v]
            _random.shuffle(_v)
            _v = "".join(_v)
            
        elif isinstance(v, _ab_mod.AbroadInitializer):
            _v = +v
            _random.shuffle(_v)
            
        elif isinstance(v, (_tc.Sequence, _tc.Uniqual)):
            _v = [e for e in v]
            _random.shuffle(_v)
            
        elif isinstance(v, _tc.Mapping):
            _v2 = [k for k in v]
            _random.shuffle(_v2)
            _v = {k: v[k] for k in _v2}
        
        else:
            error = TypeError("expected a string, mapping, sequence or result of abroad() function")
            raise error
            
        return _v
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def reverse(self, v: str) -> str: ...
    
    @classmethod
    @_tc.overload
    def reverse(self, v: _ab_mod.AbroadInitializer) -> list[int]: ...
    
    @classmethod
    @_tc.overload
    def reverse(self, v: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]]) -> list[_T]: ...
    
    @classmethod
    @_tc.overload
    def reverse(self, v: _tc.Mapping[_KT, _VT]) -> dict[_KT, _VT]: ...
    
    @classmethod
    def reverse(self, v):
        """
        \\@since 0.3.26rc2
        ```ts
        "class method" in class Tense
        ```
        Reverse a string, mapping or sequence, and return reversed iterable \\
        without changing passed iterable.
        """
        
        # note i dont really rely on reversed builtin xd
        if isinstance(v, str):
            
            _v = [c for c in v]
            _v.reverse()
            _v = "".join(_v)
            
        elif isinstance(v, _ab_mod.AbroadInitializer):
            
            _v = +v
            _v.reverse()
            
        elif isinstance(v, (_tc.Sequence, _tc.Uniqual)):
            
            _v = [e for e in v]
            _v.reverse()
            
        elif isinstance(v, _tc.Mapping):
            
            _v2 = [k for k in v]
            _v2.reverse()
            _v = {k: v[k] for k in _v2}
            
        else:
            error = TypeError("expected a string, mapping, sequence or result of abroad() function")
            raise error
        
        return _v
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def occurrences(self, v: str, *items: str, mode: _lit["case_sensitive"] = "case_sensitive") -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: str, *items: str, mode: _lit["case_insensitive"]) -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: _ab_mod.AbroadInitializer, *items: int, mode: _lit["normal"] = "normal") -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: _ab_mod.AbroadInitializer, *items: int, mode: _lit["absolute"]) -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], *items: _T) -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: _tc.Mapping[_KT, _VT], *items: _KT, mode: _lit["key"] = "key") -> int: ...
    
    @classmethod
    @_tc.overload
    def occurrences(self, v: _tc.Mapping[_KT, _VT], *items: _VT, mode: _lit["value"]) -> int: ...
    
    @classmethod
    def occurrences(self, v, *items, mode = "case_sensitive"):
        """
        \\@since 0.3.32
        
        Returns number of how many times specified item appears \\
        or items appear in a sequence
        """
        
        o = 0
        m = mode.lower()
        
        if not isinstance(m, str):
            
            error = TypeError("parameter 'mode' is not a string")
            raise error
        
        if m not in ("case_sensitive", "case_insensitive", "normal", "absolute", "key", "value"):
            
            error = ValueError("parameter 'mode' provides invalid mode")
            raise error
        
        if isinstance(v, (str, _ab_mod.AbroadInitializer, _tc.Sequence, _tc.Uniqual, _tc.Mapping)) and reckon(v) == 0:
            return 0
        
        if isinstance(v, str):
            
            if m not in ("case_sensitive", "case_insensitive"):
                
                error = ValueError("for strings parameter 'mode' can take one of 2 modes: 'case_sensitive' and 'case_insensitive'")
                raise error
            
            
            _v = v.split()
            _s = list(items)
            
            if not self.isList(_s, str):
                
                error = ValueError("parameter 'items' doesn't utterly consist of string items")
                raise error
            
            if m == "case_insensitive":
                _s = [s.lower() for s in items]
                
            for s in _v:
                
                if (s in _s and m == "case_sensitive") or (s.lower() in _s and m == "case_insensitive"):
                    o += 1
                    
            return o
                    
        elif isinstance(v, _ab_mod.AbroadInitializer):
            
            if m not in ("normal", "absolute"):
                
                error = ValueError("for instances of internal class being result of abroad() function, parameter 'mode' can take one of 2 modes: 'normal' and 'absolute'".format(_ab_mod.AbroadInitializer.__name__))
                raise error
            
            _i = list(items)
            
            if not self.isList(_i, int):
                
                error = ValueError("parameter 'items' doesn't utterly consist of integer items")
                raise error
            
            if m == "absolute":
                _i = [abs(i) for i in items]
                
                
            for i in v:
                
                if (i in _i and m == "normal") or (abs(i) in _i and m == "absolute"):
                    o += 1
                    
            return o
                    
        elif isinstance(v, (_tc.Sequence, _tc.Uniqual)):
            
            _v = list(v)
            
            for e in _v:
                
                if e in items:
                    o += 1
                    
            return o
                    
        elif isinstance(v, _tc.Mapping):
            
            if m not in ("key", "value"):
                
                error = ValueError("for mappings parameter 'mode' can take one of 2 modes: 'key' and 'value'")
                raise error
            
            if m == "key":
                _v = list(v.keys())
                
            else:
                _v = list(v.values())
                
            for e in _v:
                
                if e in items:
                    o += 1
                    
            return o
            
        else:
            error = TypeError("expected a string, mapping, sequence or result of abroad() function")
            raise error
                    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def difference(self, v1: type[_T1], v2: type[_T2], /, invert: bool = False, value_check: bool = True) -> list[str]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: type[_T1], v2: _tc.Mapping[str, _T2], /, invert: bool = False, value_check: bool = True) -> list[str]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _tc.Mapping[str, _T1], v2: type[_T2], /, invert: bool = False, value_check: bool = True) -> list[str]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _tc.Mapping[_KT1, _VT1], v2: _tc.Mapping[_KT2, _VT2], /, invert: _lit[False] = False, value_check: bool = True) -> list[_KT1]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _tc.Mapping[_KT1, _VT1], v2: _tc.Mapping[_KT2, _VT2], /, invert: _lit[True], value_check: bool = True) -> list[_KT2]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: type[_T], v2: _uni[_tc.Sequence[str], _tc.Uniqual[str]], /, invert: bool = False, value_check: bool = True) -> list[str]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _uni[_tc.Sequence[str], _tc.Uniqual[str]], v2: type[_T], /, invert: bool = False, value_check: bool = True) -> list[str]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _ab_mod.AbroadInitializer, v2: _uni[_tc.Sequence[int], _tc.Uniqual[int]], /, invert: bool = False) -> list[int]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _uni[_tc.Sequence[int], _tc.Uniqual[int]], v2: _ab_mod.AbroadInitializer, /, invert: bool = False) -> list[int]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _ab_mod.AbroadInitializer, v2: _ab_mod.AbroadInitializer, /, invert: bool = False) -> list[int]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _uni[_tc.Sequence[_T1], _tc.Uniqual[_T1]], v2: _uni[_tc.Sequence[_T2], _tc.Uniqual[_T2]], /, invert: _lit[False] = False) -> list[_T1]: ...
    
    @classmethod
    @_tc.overload
    def difference(self, v1: _uni[_tc.Sequence[_T1], _tc.Uniqual[_T1]], v2: _uni[_tc.Sequence[_T2], _tc.Uniqual[_T2]], /, invert: _lit[True]) -> list[_T2]: ...
    
    @classmethod
    def difference(self, v1, v2, /, invert = False, value_check = True):
        """
        \\@since 0.3.32
        
        Find items, which belong to `v1`, but do not belong to `v2` (math difference `v1 \\ v2`). \\
        With `invert` being `True` it is vice versa (`v2 \\ v1`).
        
        When `value_check` is `True`, values from fields in classes or dictionairies will be also checked. \\
        In this case when values are different, then they are included in the returned list.
        
        For class fields with roundly underscored names, these fields aren't counted.
        """
        _v1, _v2 = [{} for _ in abroad(2)]
        
        if isinstance(v1, type):
            
            if isinstance(v2, type):
                
                _v1 = {k: v1.__annotations__[k] for k in v1.__annotations__ if k[:1] != "_"}
                _v2 = {k: v2.__annotations__[k] for k in v2.__annotations__ if k[:1] != "_"}
            
            elif isinstance(v2, _tc.Mapping):
                
                if not self.isList([k for k in v2], str):
                    
                    error = ValueError("with comparison with a class expected a string-key-typed mapping")
                    raise error
                
                _v1 = {k: v1.__annotations__[k] for k in v1.__annotations__ if k[:1] != "_"}
                _v2 = {k: v2[k] for k in v2 if k[:1] != "_"}
            
            elif isinstance(v2, (_tc.Sequence, _tc.Uniqual)):
                
                if not self.isList([k for k in v2], str):
                    
                    error = ValueError("with comparison with a class expected a string-typed sequence")
                    raise error
                
                _v1 = [k for k in v1.__annotations__]
                _v2 = list(v2)
                
            else:
                error = TypeError("with comparison with a class expected another class or string sequence, mapping with string keys")
                raise error
        
        elif isinstance(v1, _tc.Mapping):
            
            if isinstance(v2, type):
                
                if not self.isList([k for k in v1], str):
                    
                    error = ValueError("with comparison with a class expected a string-key-typed mapping")
                    raise error
                
                _v1 = {k: v1[k] for k in v1 if k[:1] != "_"}
                _v2 = {k: v2.__annotations__[k] for k in v2.__annotations__ if k[:1] != "_"}
                
            elif isinstance(v2, _tc.Mapping):
                
                _v1 = {k: v1[k] for k in v1 if k[:1] != "_"}
                _v2 = {k: v2[k] for k in v2 if k[:1] != "_"}
                
            else:
                error = TypeError("with comparison with a mapping expected another mapping or a class")
                raise error
        
        elif isinstance(v1, _ab_mod.AbroadInitializer):
            
            if isinstance(v2, _ab_mod.AbroadInitializer):
                
                _v1 = +v1
                _v2 = +v2
                
                if invert:
                    return [e for e in _v1 if e not in _v2]
                
                else:
                    return [e for e in _v2 if e not in _v1]
                
            elif isinstance(v2, (_tc.Sequence, _tc.Uniqual)):
                
                if not self.isList([e for e in v2], int):
                    
                    error = ValueError("with comparison with a result from abroad() function expected an integer sequence")
                    raise error
                
                _v1 = +v1
                _v2 = list(v2)
                
            else:
                error = TypeError("with comparison with a result from abroad() function expected another abroad() function result or integer sequence")
                raise error
            
        elif isinstance(v1, (_tc.Sequence, _tc.Uniqual)):
            
            if isinstance(v2, type):
                
                if not self.isList([k for k in v1], str):
                    
                    error = ValueError("with comparison with a class expected a string-typed sequence")
                    raise error
                
                _v1 = list(v1)
                _v2 = [k for k in v2.__annotations__]
                
            elif isinstance(v2, _ab_mod.AbroadInitializer):
                
                if not self.isList([e for e in v2], int):
                    
                    error = ValueError("with comparison with a result from abroad() function expected an integer sequence")
                    raise error
                
                _v1 = list(v1)
                _v2 = +v2
                
            elif isinstance(v2, (_tc.Sequence, _tc.Uniqual)):
                
                _v1 = list(v1)
                _v2 = list(v2)
                
        if invert:
            _v1, _v2 = _v2, _v1
            
        _res = [k for k in _v1 if k not in _v2]
        
        if value_check and isinstance(_v1, dict) and isinstance(_v2, dict):
            
            for k in _v1:
                
                if k in _v2 and _v1[k] != _v2[k]:
                    _res.append(k)
        
        _res.sort()
        return _res
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def intersection(self, v: int, /, *_: int) -> int: ...
    
    @classmethod
    @_tc.overload
    def intersection(self, v: bool, /, *_: bool) -> bool: ...
    
    @classmethod
    @_tc.overload
    def intersection(self, v: _tc.Iterable[_T], /, *_: _tc.Iterable[_Any]) -> list[_T]: ...
                        
    @classmethod
    def intersection(self, v, /, *_):
        """
        \\@since 0.3.34
        
        Returns list of items, which appear in all sequences (math intersection `v1 \u2229 v2 \u2229 ... \u2229 vN`). \\
        In case of integers returned is bitwise AND (`&`) taken on all integers. \\
        In case of boolean values returned is logical AND (`and`) taken on all boolean values.
        """
        if isinstance(v, bool):
            
            if reckon(_) == 0:
                return v
            
            else:
                
                if not self.isList([e for e in _], bool):
                    
                    error = ValueError("expected every value a boolean value")
                    raise error
                
                _v = v
                
                for e in _:
                    _v = _v and e
                    
                return _v
        
        elif isinstance(v, int):
            
            if reckon(_) == 0:
                return v
            
            else:
                
                if not self.isList([e for e in _], int):
                    
                    error = ValueError("expected every value an integer")
                    raise error
                
                _v = v
                
                for e in _:
                    _v &= e
                    
                return _v
            
        elif isinstance(v, _tc.Iterable):
            
            if reckon(_) == 0:
                return list(v)
            
            else:
                
                if not self.isList([e for e in _], _tc.Iterable):
                    
                    error = ValueError("expected every value an iterable")
                    raise error
                
                return list(set(v).intersection(*_))
            
        else:
            error = TypeError("expected every value either integers, boolean values or iterable objects")
            raise error
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def union(self, v: int, /, *_: int) -> int: ...
    
    @classmethod
    @_tc.overload
    def union(self, v: bool, /, *_: bool) -> bool: ...
    
    @classmethod
    @_tc.overload
    def union(self, v: _tc.Iterable[_T1], /, *_: _tc.Iterable[_T2]) -> list[_uni[_T1, _T2]]: ...
    
    @classmethod
    def union(self, v, /, *_):
        """
        \\@since 0.3.34
        
        Returns list of items, which appear in any sequences (math union `v1 \u222A v2 \u222A ... \u222A vN`). \\
        In case of integers returned is bitwise OR (`|`) taken on all integers. \\
        In case of boolean values returned is logical OR (`or`) taken on all boolean values.
        """
        
        if isinstance(v, bool):
            
            if reckon(_) == 0:
                return v
            
            else:
                
                if not self.isList([e for e in _], bool):
                    
                    error = ValueError("expected every value a boolean value")
                    raise error
                
                _v = v
                
                for e in _:
                    _v = _v or e
                    
                return _v
        
        elif isinstance(v, int):
            
            if reckon(_) == 0:
                return v
            
            else:
                
                if not self.isList([e for e in _], int):
                    
                    error = ValueError("expected every value an integer")
                    raise error
                
                _v = v
                
                for e in _:
                    _v |= e
                    
                return _v
            
        elif isinstance(v, _tc.Iterable):
            
            if reckon(_) == 0:
                return list(v)
            
            else:
                
                if not self.isList([e for e in _], _tc.Iterable):
                    
                    error = ValueError("expected every value an iterable")
                    raise error
                
                return list(set(v).union(*_))
            
        else:
            error = TypeError("expected every value either integers, boolean values or iterable objects")
            raise error
    
    @classmethod
    def append(self, i: _tc.Iterable[_T], /, *items: _T):
        """
        \\@since 0.3.27a4
        ```ts
        "class method" in class Tense
        ```
        Same as `list.append()`, just variable amount of items can be passed. \\
        Input list remains non-modified, and returned is its modified copy.
        
        Since 0.3.34, because mutable sequences are normally coerced to list, allowed are also \\
        all other iterables
        """
        return [e for e in i] + [e for e in items]
    
    @classmethod
    def extend(self, i: _tc.Iterable[_T], /, *iters: _tc.Iterable[_T]):
        """
        \\@since 0.3.27a4
        ```ts
        "class method" in class Tense
        ```
        Same as `list.extend()`, just variable amount of iterables can be passed. \\
        Input list remains non-modified, and returned is its modified copy.
        
        Since 0.3.34, because mutable sequences are normally coerced to list, allowed are also \\
        all other iterables
        """
        _seq = [e for e in i]
        
        for e in iters:
            _seq.extend(e)
                
        return _seq
    
    @classmethod
    def exclude(self, i: _tc.Iterable[_T], /, *items: _T):
        """
        \\@since 0.3.34
        
        Return a new list from iterable without items specified
        """
        return [e for e in i if e not in items]
    
    @classmethod
    def explode(self, s: str, /, separator: _opt[str] = None, max = -1, no_empty = False):
        """
        \\@since 0.3.34
        
        Reference from PHP inbuilt function `explode()`. Split a string \\
        using specified separator into a string list.
        """
        _params = [k for k in self.explode.__annotations__ if k not in ("self", "return")]
        
        # error message template
        _msgtmpl = "expected {} in parameter '{}'"
        
        _msg = ["" for _ in abroad(_params)]
        _msg[0] = _msgtmpl.format("a string", _params[0])
        _msg[1] = _msgtmpl.format("a string or None", _params[1])
        _msg[2] = _msgtmpl.format("an integer for -1 above (not being zero)", _params[2])
        
        if not self.isString(s):
            error = TypeError(_msg[0])
            raise error
        
        if not self.isString(separator) and not self.isNone(separator):
            error = TypeError(_msg[1])
            raise error
        
        if not self.isInteger(max) or (self.isInteger(max) and max != -1 and max < 1):
            error = TypeError(_msg[2])
            raise error
        
        if no_empty:
            return [k for k in s.split(separator, max) if reckon(k) != 0]
        
        return s.split(separator, max)

    @classmethod
    def clear(self, *v: _tc.Union[_tc.MutableSequence[_T], str]):
        """
        \\@since 0.3.27a4
        ```ts
        "class method" in class Tense
        ```
        Same as `list.clear()`, just variable amount of lists can be passed. \\
        Since 0.3.27b1 strings are also passable.
        """
        for seq in v:
            
            if self.isString(seq):
                seq = ""
                
            else:
                seq.clear()
                
    @classmethod
    def copy(self, x: _T):
        """\\@since 0.3.34"""
        return _copy.copy(x)
    
    
    @classmethod
    def deepcopy(self, x: _T, memo: _opt[dict[int, _Any]] = None, _nil: _Any = []):
        """\\@since 0.3.34"""
        return _copy.deepcopy(x, memo, _nil)
    

    # OVERLOAD 0.3.34
    if False:
        @classmethod
        def eval(self, source: _tc.Union[str, _tc.Buffer, _ast.Module, _ast.Expression, _ast.Interactive], ast = False):
            
            if not ast:
                return eval(compile(source))
            
            else:
                
                if not self.isString(source) and not isinstance(source, _tc.Buffer):
                    error = TypeError("for ast.literal_eval expected a string or buffer")
                    raise error
                
                else:
                    return _ast.literal_eval(str(source) if not self.isString(source) else source)
                
    else:
        
        # for ValueError exception reason, there is no 'str' in this overload, instead create an AST instance
        @classmethod
        @_tc.overload
        def eval(self, source: _ast.AST, /) -> _Any:
            """\\@since 0.3.26rc2"""
            ...
            
        @classmethod
        @_tc.overload
        def eval(self, source: _uni[str, _tc.Buffer, _tc.CodeType], /, globals: _opt[dict[str, _Any]] = None, locals: _opt[_tc.Mapping[str, object]] = None) -> _Any:
            """\\@since 0.3.26rc2"""
            ...
            
        @classmethod
        @_tc.overload
        def eval(
            self,
            source: tuple[_uni[str, _tc.Buffer, _ast.Module, _ast.Expression, _ast.Interactive], _uni[str, _tc.Buffer, _tc.PathLike[_Any]], str],
            /,
            globals: _opt[dict[str, _Any]] = None,
            locals: _opt[_tc.Mapping[str, object]] = None,
            flags: int = 0,
            dont_inherit: bool = False,
            optimize: int = -1
        ) -> _Any:
            """\\@since 0.3.26rc2"""
            ...
            
        @classmethod
        def eval(self, source, /, globals = None, locals = None, flags = 0, dont_inherit = False, optimize = -1):
            
            if (
                isinstance(source, (str, _ast.AST, _tc.Buffer, _tc.CodeType)) and
                (isinstance(globals, dict) or globals is None) and
                (isinstance(locals, _tc.Mapping) or locals is None)
            ):
                
                if isinstance(source, (str, _tc.Buffer, _tc.CodeType)):
                            
                    return _builtins.eval(source, globals, locals)
                
                else:
                    
                    return _ast.literal_eval(source)
                    
            elif isinstance(source, tuple):
                
                if reckon(source) == 3 and (
                    isinstance(source[0], (str, _tc.Buffer, _ast.Module, _ast.Expression, _ast.Interactive)) and # 'source'
                    isinstance(source[1], (str, _tc.Buffer, _os.PathLike)) and # 'fn': importing 'PathLike' from 'os' since from abc module in this project is a type alias
                    isinstance(source[2], str) and # 'mode'
                    (isinstance(globals, dict) or globals is None) and
                    (isinstance(locals, _tc.Mapping) or locals is None) and
                    isinstance(flags, int) and
                    isinstance(dont_inherit, bool) and
                    isinstance(optimize, int)
                ):
                    
                    _res = compile(source[0], source[1], source[2], flags, dont_inherit, optimize)
                    
                    if isinstance(_res, _ast.AST):
                        
                        return _ast.literal_eval(_res)
                    
                    elif isinstance(_res, str):
                        
                        try:
                            return _ast.literal_eval(_res)
                        
                        except ValueError:
                            return _builtins.eval(_res, globals, locals)
                    
                    elif isinstance(_res, _tc.CodeType):
                        return _builtins.eval(_res, globals, locals)
                    
                    return _res # I would care less about this result
            
            error = TypeError("no matching function signature; make sure every type requirement is satisfied")
            raise error
                        
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def bisect(self, a: _tc.SizeableItemGetter[_T_richComparable], x: _T_richComparable, /, l: int = 0, h: _opt[int] = None, dir: _BisectMode = BISECT_RIGHT, *, key: None = None) -> int:
        """\\@since 0.3.26rc2"""
        ...
    
    @classmethod
    @_tc.overload
    def bisect(self, a: _tc.SizeableItemGetter[_T], x: _T_richComparable, /, l: int = 0, h: _opt[int] = None, dir: _BisectMode = BISECT_RIGHT, *, key: _cal[[_T], _T_richComparable]) -> int:
        """\\@since 0.3.26rc2"""
        ...
            
    @classmethod
    def bisect(self, a, x, /, l = 0, h = None, dir = BISECT_RIGHT, *, key = None):
        
        if (
            self.isInteger(l) and
            (self.isInteger(h) or self.isNone(h)) and
            isinstance(dir, _BisectMode)
        ):
            
            if not _inspect.isfunction(key):
                return _bisect.bisect_right(a, x, l, h) if dir == self.BISECT_RIGHT else _bisect.bisect_left(a, x, l, h)
            
            else:
                return _bisect.bisect_right(a, x, l, h, key = key) if dir == self.BISECT_RIGHT else _bisect.bisect_left(a, x, l, h, key = key)
        
        error = TypeError("no matching function signature; make sure every type requirement is satisfied")
        raise error
    
    # OVERLOAD 0.3.34
    if False:
        @classmethod
        def insort(self, seq: _tc.MutableSequence[_T], item: _T, low: int = 0, high: _opt[int] = None, mode: _InsortMode = INSORT_RIGHT, key: _opt[_cal[[_T], _T_richComparable]] = None):
            
            _seq = seq
            if mode == self.INSORT_LEFT:
                _bisect.insort_left(_seq, item, low, high, key = key)
                
            elif mode == self.INSORT_RIGHT:
                _bisect.insort_right(_seq, item, low, high, key = key)
                
            else:
                error = TypeError("incompatible value for 'mode' parameter. Expected one of constants: 'INSORT_LEFT' or 'INSORT_RIGHT'")
                raise error
            return _seq
        
    else:
        
        @classmethod
        @_tc.overload
        def insort(self, a: _tc.MutableSequence[_T_richComparable], x: _T_richComparable, /, l: int = 0, h: _opt[int] = None, dir: _InsortMode = INSORT_RIGHT, *, key: None = None) -> None:
            """\\@since 0.3.26rc2"""
            ...
            
        @classmethod
        @_tc.overload
        def insort(self, a: _tc.MutableSequence[_T], x: _T_richComparable, /, l: int = 0, h: _opt[int] = None, dir: _InsortMode = INSORT_RIGHT, *, key: _cal[[_T], _T_richComparable]) -> None:
            """\\@since 0.3.26rc2"""
            ...
            
        @classmethod
        def insort(self, a, x, /, l = 0, h = None, dir = INSORT_RIGHT, *, key = None):
            
            if (
                self.isInteger(l) and
                (self.isInteger(h) or self.isNone(h)) and
                isinstance(dir, _InsortMode)
            ):
                
                if not _inspect.isfunction(key):
                    return _bisect.insort_right(a, x, l, h) if dir == self.INSORT_RIGHT else _bisect.insort_left(a, x, l, h)
                
                else:
                    return _bisect.insort_right(a, x, l, h, key = key) if dir == self.INSORT_RIGHT else _bisect.insort_left(a, x, l, h, key = key)
            
            error = TypeError("no matching function signature; make sure every type requirement is satisfied")
            raise error
    
    @classmethod
    def print(self, *values: object, separator: _opt[str] = " ", ending: _opt[str] = "\n", file: _uni[_tc.Writable[str], _tc.Flushable, None] = None, flush: bool = False, invokeAs = "INSERTION"):
        """
        \\@since 0.3.25
        ```
        "class method" in class Tense
        ```
        Same as `print()`, just with `INSERTION` beginning. It can be \\
        changed with `invokeAs` parameter. Since 0.3.26a1 this method \\
        returns reference to this class. Since 0.3.27b1, if setting \\
        `TenseOptions.insertionMessage` was `False`, `invokeAs` \\
        parameter will lose its meaning.
        """
        if TenseOptions.insertionMessage is True:
            e = super().fencordFormat()
            print(f"\33[1;90m{e}\33[1;38;5;45m {invokeAs}\33[0m", *values, sep = separator, end = ending, file = file, flush = flush)
            
        else:
            print(*values, sep = separator, end = ending, file = file, flush = flush)
            
        return self
    
    @classmethod
    @_tc.overload
    def random(self, x: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /) -> _T: ...
    
    @classmethod
    @_tc.overload
    def random(self, x: int, /) -> int: ...
    
    @classmethod
    @_tc.overload
    def random(self, x: int, y: int, /) -> int: ...
    
    @classmethod
    def random(self, x, y = None, /):
        """
        \\@since 0.3.24 (standard since 0.3.25) \\
        \\@lifetime ≥ 0.3.24 \\
        \\@modified 0.3.25, 0.3.26rc2 (support for `tkinter.IntVar`), 0.3.31 (cancelled support for `tkinter.IntVar`) \\
        0.3.34 (overloads)
        ```
        "class method" in class Tense
        ```
        With one parameter, returns an item from a sequence or integer from range [0, x) \\
        With two parameters, returns an integer in specified range [x, y] or [y, x] if x > y.
        """
        if self.isNone(y):
            
            if self.isInteger(x):
                
                if x <= 1:
                    error = ValueError("expected a positive integer above 1")
                    raise error
                
                # note random.randrange() is also a thing
                return _secrets.randbelow(x)
            
            elif isinstance(x, (_tc.Sequence, _tc.Uniqual)):
                
                return _random.choice(list(x))
            
        a = [x, y]
        
        if self.isList(a, int):
                
            _x, _y = x, y
            
            if x > y:
                _x, _y = _y, _x
            
            return _random.randint(_x, _y)
            
        error = TypeError("no matching function signature")
        raise error
                
    @classmethod
    def randstr(self, lower = True, upper = True, digits = True, special = True, length = 10):
        """
        \\@since 0.3.9 \\
        \\@lifetime ≥ 0.3.9; < 0.3.24; ≥ 0.3.25 \\
        to 0.3.34 known as `NennaiRandomize.randomizeStr()`
        ```
        # created 15.07.2024
        "class method" in class NennaiRandomize
        ```
        - `lower` - determine, if you want to include all lowercased letters from english alphabet. Defaults to `True`
        - `upper` - determine, if you want to include all uppercased letters from english alphabet. Defaults to `True`
        - `digits` - determine, if you want to include all numbers. Defaults to `True`
        - `special` - determine, if you want to include all remaining chars accessible normally via English keyboard. \\
        Defaults to `True`
        - `length` - allows to specify the length of returned string. Defaults to `10`.
        """
        # code change 0.3.34
        conv = [""]
        conv.clear()
        ret = ""
        _ALPHA = _string.ascii_lowercase # abcdefghijklmnopqrstuvwxyz
        _DIGITS = _string.digits # 0123456789
        _SPECIAL = r"""`~!@#$%^&*()-_=+[]{};:'"\|,.<>/?""" # gimmick from 'string' module
        
        if lower:
            conv.extend([e for e in _ALPHA])
                
        if upper:
            conv.extend([e for e in _ALPHA.upper()])
                
        if digits:
            conv.extend([e for e in _DIGITS])
                
        if special:
            conv.extend([e for e in _SPECIAL])
        
        # there no matter if negative or positive
        for _ in abroad(length):
            ret += self.pick(conv)
            
        return ret
    
    @classmethod
    def uuidPrimary(self, node: _uni[int, None] = None, clockSeq: _uni[int, None] = None):
        """
        \\@since 0.3.26a1 \\
        \\@modified 0.3.31 (cancelled support for `tkinter.IntVar`)
        ```ts
        // created 20.07.2024
        "class method" in class Tense
        ```
        Return an UUID from host ID, sequence number and the current time.
        """
        _n = node
        _c = clockSeq
        return _uuid.uuid1(node = _n, clock_seq = _c)
    
    @classmethod
    def uuidMd5(self, namespace: _tc.UUID, name: _uni[str, bytes]):
        """
        \\@since 0.3.26a1 \\
        \\@modified 0.3.31 (cancelled support for `tkinter.StringVar`)
        ```ts
        // created 20.07.2024
        "class method" in class Tense
        ```
        Return an UUID from the MD5 (Message Digest) hash of a namespace UUID and a name
        """
        return _uuid.uuid3(namespace = namespace, name = name)
    
    @classmethod
    def uuidRandom(self):
        """
        \\@since 0.3.26a1
        ```ts
        // created 20.07.2024
        "class method" in class Tense
        ```
        Return a random UUID
        """
        return _uuid.uuid4()
    
    @classmethod
    def uuidSha1(self, namespace: _tc.UUID, name: _uni[str, bytes]):
        """
        \\@since 0.3.26a1 \\
        \\@modified 0.3.31 (cancelled support for `tkinter.StringVar`)
        ```ts
        "class method" in class Tense
        ```
        Return an UUID from the SHA-1 (Secure Hash Algorithm) hash of a namespace UUID and a name
        """
        return _uuid.uuid5(namespace = namespace, name = name)
    
    @classmethod
    @_tc.overload
    def pick(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, secure: bool = False) -> _T: ... # sets act the same as lists in this case
    
    @classmethod
    @_tc.overload
    def pick(self, i: _tc.Mapping[_KT, _VT], /, secure: bool = False) -> _VT: ...
    
    @classmethod
    def pick(self, i, /, secure = False):
        """
        \\@since 0.3.8 (standard since 0.3.24) \\
        \\@lifetime ≥ 0.3.8 \\
        \\@modified 0.3.25, 0.3.26rc2, 0.3.26rc3, 0.3.34
        ```ts
        "class method" in class Tense
        ```
        Returns random item from a sequence (+ mapping since 0.3.34)
        """
        if isinstance(i, (_tc.Sequence, _tc.Uniqual)):
            
            if secure:
                return _secrets.choice(i)
            
            else:
                return _random.choice(i)
        
        elif isinstance(i, _tc.Mapping):
            
            if secure:
                return i[_secrets.choice([k for k in i])]
            
            else:
                return i[_random.choice([k for k in i])]
        
        else:
            error = TypeError("expected a sequence or mapping")
            raise error
    
    if False: # deprecated since 0.3.25, removed 0.3.31
        def error(self, handler: type[Exception], message: _uni[str, None] = None):
            """
            \\@since 0.3.24 \\
            \\@deprecated 0.3.25
            ```
            "class method" in class Tense
            ```
            """
            _user_defined_error = handler
            _user_defined_reason = message
            if _user_defined_reason is None:
                raise _user_defined_error()
            else:
                raise _user_defined_error(_user_defined_reason)
            
            
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def first(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: None = None, default: _S = None) -> _uni[_S, _T]: ...
    
    @classmethod
    @_tc.overload
    def first(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: _cal[[_T], bool], default: _S = None) -> _uni[_S, _T]: ...
            
    @classmethod
    def first(self, i, /, condition = None, default = None):
        """
        \\@since 0.3.26rc2
        ```
        "class method" in class Tense
        ```
        Return first element in a `seq` (sequence) which satisfies `condition`. If none found, returned \\
        is default value defined via parameter `default`, which by default has value `None`. On 0.3.27a4 \\
        removed this parameter, it has been restored on 0.3.34
        """
        
        if not _is_condition_callback(condition) and not self.isNone(condition):
            error = TypeError("expected 'condition' parameter to be a callable or 'None'")
            raise error
        
        _seq = list(i)
        
        for _i in abroad(_seq):
            
            if condition is not None and condition(_seq[_i]):
                return _seq[_i]
            
            else:
                if _seq[_i]: return _seq[_i]
                
        return default
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def last(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: None = None, default: _S = None) -> _uni[_S, _T]: ...
    
    @classmethod
    @_tc.overload
    def last(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: _cal[[_T], bool], default: _S = None) -> _uni[_S, _T]: ...
    
    @classmethod
    def last(self, i, /, condition = None, default = None):
        """
        \\@since 0.3.26rc2
        ```
        "class method" in class Tense
        ```
        Return last element in a `seq` (sequence) which satisfies `condition`. If none found, returned is default \\
        value defined via parameter `default`, which by default has value `None`. On 0.3.27a4 removed this parameter, \\
        it has been restored on 0.3.34
        """
        if not _is_condition_callback(condition) and not self.isNone(condition):
            error = TypeError("expected 'condition' parameter to be a callable or 'None'")
            raise error
        
        _seq = list(i)
        
        for _i in self.abroadNegative(1, _seq):
            
            if condition is not None and condition(_seq[_i]):
                return _seq[_i]
            
            else:
                if _seq[_i]: return _seq[_i]
                
        return default
    
    # OVERLOAD 0.3.34
    @classmethod
    @_tc.overload
    def any(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: None = None) -> bool: ...
    
    @classmethod
    @_tc.overload
    def any(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: _cal[[_T], bool]) -> bool: ...
    
    @classmethod
    def any(self, i, /, condition = None):
        """
        \\@since 0.3.26rc2
        ```
        "class method" in class Tense
        ```
        Equivalent to `any()` in-built function, but this method returns list of items, which satisfied `condition`, \\
        which has default value `None`. If none found, returned is empty list.
        Change 0.3.34: `any()` now returns boolean rather than shallow copy of sequence
        """
        if not _is_condition_callback(condition) and not self.isNone(condition):
            error = TypeError("expected 'condition' parameter to be a callable or 'None'")
            raise error
        
        for e in list(i):
            
            if (condition is not None and condition(e)) or (condition is None and bool(e)):
                return True
            
        return False
            
    @classmethod
    @_tc.overload
    def all(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: None = None) -> bool: ...
    
    @classmethod
    @_tc.overload
    def all(self, i: _uni[_tc.Sequence[_T], _tc.Uniqual[_T]], /, condition: _cal[[_T], bool]) -> bool: ...
    
    @classmethod
    def all(self, i, /, condition = None): # slash was after 'condition' parameter before (0.3.34)
        """
        \\@since 0.3.26rc2
        ```
        "class method" in class Tense
        ```
        Change 0.3.34: `all()` now returns boolean rather than shallow copy of sequence
        Change 0.3.27a4: removed parameter `default`.
        """
        if not _is_condition_callback(condition) and not self.isNone(condition):
            error = TypeError("expected 'condition' parameter to be a callable or 'None'")
            raise error
        
        _placeholder = True
        
        for e in list(i):
            
            if condition is not None:
                _placeholder = _placeholder and condition(e)
                
            else:
                _placeholder = _placeholder and bool(e)
        
        return _placeholder
    
    @classmethod
    def probability2(self, x: _T = 1, y: _T = 0, frequency = 1, length: _ProbabilityLengthType = 10000):
        """
        \\@since 0.3.8 (standard since 0.3.9) \\
        \\@lifetime ≥ 0.3.8; < 0.3.24; ≥ 0.3.25 \\
        \\@modified 0.3.26a3, 0.3.26rc1, 0.3.31 \\
        https://aveyzan.glitch.me/tense/py/method.probability.html#2
        ```ts
        "class method" in class Tense
        ``` \n
        ``` \n
        # syntax since 0.3.25
        def probability2(x = 1, y = 0, frequency = 1, length = 10000): ...
        # syntax for 0.3.19 - 0.3.23; on 0.3.19 renamed from probability()
        def probability2(rareValue = 1, usualValue = 0, frequency = 1, length = 10000): ...
        # syntax before 0.3.19
        def probability(value = 1, frequency = 1, length = 10000): ...
        ```
        Randomize a value using probability `frequency/length` applied on parameter `x`. \\
        Probability for parameter `y` will be equal `(length - frequency)/length`. \\
        Default values:
        - for `x`: 1
        - for `y`: 0
        - for `frequency`: 1
        - for `length`: 10000 (since 0.3.26a3 `length` can also have value `-1`)

        To be more explanatory, `x` has `1/10000` chance to be returned by default (in percents: 0.01%), \\
        meanwhile the rest chance goes to `y` (`9999/10000`, 99.99%), hence `y` will be returned more \\
        frequently than `x`. Exceptions:
        - for `frequency` equal 0 or `x` equal `y`, returned is `y`
        - for `frequency` greater than (or since 0.3.25 equal) `length` returned is `x`
        """
        
        # due to swap from IntegerEnum to Enum in _ProbabilityLength class's subclass
        # it had to be replaced
        _length = 10000 if length in (-1, self.PROBABILITY_COMPUTE) else length
        _frequency = frequency
        
        # 0.3.33: refraining from using string literals, since they will need to be manually changed
        # once parameter names are changed
        # note that 'return' keyword is reserved for return annotation with '->' operator
        _params = [k for k in self.probability2.__annotations__ if k != "return"]
        _options = [k for k in TenseOptions.__dict__ if k[:1] != "_"]
            
        if not self.isInteger(_frequency):
            error = TypeError("expected an integer in parameter '{}'".format(_params[2]))
            raise error
        
        elif _frequency < 0:
            error = ValueError("expected a non-negative integer in parameter '{}'".format(_params[2]))
            raise error
        
        # types must match, otherwise you can meet an union-typed result, which is not useful during
        # type inspection, since you need to append appropriate 'if' statement!
        # exception: a function result being a union-typed one
        if not self.isList([x, y], type(x)):
            error = TypeError("provided types in parameters '{}' and '{}' do not match".format(_params[0], _params[1]))
            raise error
        
        if not self.isInteger(length) and length != self.PROBABILITY_COMPUTE:
            error = TypeError("expected an integer or constant '{}' in parameter '{}'".format(self.PROBABILITY_COMPUTE.name, _params[3]))
            raise error
        
        elif _length == 0:
            error = ZeroDivisionError("expected integer value from -1 or above in parameter '{}', but not equal zero".format(_params[3]))
            raise error
        
        elif _length > sys.maxsize and not TenseOptions.disableProbability2LengthLimit:
            error = ValueError("integer value passed to parameter '{}' is too high, expected value below or equal {}. If you want to remove this error, toggle on option '{}'".format(
                _params[3], sys.maxsize, "{}.{}".format(TenseOptions.__name__, _options[2])
            ))
            raise error
        
        elif _length < -1:
            error = ValueError("parameter '{}' may not have a negative integer value".format(_params[3]))
            raise error
        
        # these statements are according to probability math definition
        # once 'x' and 'y' are the same, there is no reason to activate loop at all
        if x == y or _frequency == 0:
            return y
        
        if _frequency >= _length:
            return x
        
        if TenseOptions.disableProbability2LengthLimit:
            
            # 0.3.33: change in _frequency (removed minus one) so that result will be
            # displayed correctly
            r = self.random(1, _length)
            return x if r >= 1 and r <= _frequency else y
            
        else:
            # type of list will be deduced anyway, so there annotation may be skipped
            # 0.3.31: shortened code, reduced due to usage of list comprehension
            a = [y for _ in abroad(_length - _frequency)] + [x for _ in abroad(_length - _frequency, _length)]
            return self.pick(a)
    
    @classmethod
    def probability(self, *vf: _ProbabilityType[int], length: _ProbabilityLengthType = PROBABILITY_COMPUTE):
        """
        \\@since 0.3.8 (standard since 0.3.9) \\
        \\@lifetime ≥ 0.3.8 \\
        \\@modified 0.3.19, 0.3.24, 0.3.25, 0.3.26a3, 0.3.26b3, 0.3.26rc1, 0.3.26rc2, 0.3.31 \\
        https://aveyzan.glitch.me/tense/py/method.probability.html
        ```ts
        "class method" in class Tense
        ``` \n
        ``` \n
        # since 0.3.25
        def probability(*vf: int | list[int] | tuple[int, int | None] | dict[int, int | None] | deque[int], length: int = PROBABILITY_COMPUTE): ...

        # for 0.3.24
        def probability(*valuesAndFrequencies: int | list[int] | tuple[int] | dict[int, int | None], length: int = PROBABILITY_ALL): ...

        # during 0.3.19 - 0.3.23; on 0.3.19 renamed
        def probability(*valuesAndFrequencies: int | list[int] | tuple[int], length: int = -1): ...

        # during 0.3.8 - 0.3.18
        def complexity(values: list[_T] | tuple[_T], frequencies: int | list[int] | tuple[int], length: int = 10000): ...
        ```
        Extended version of `Tense.probability2()` method. Instead of only 2 values user can put more than 2. \\
        Nevertheless, comparing to the same method, it accepts integers only.

        *Parameters*:

        - `vf` - this parameter waits at least for 3 values (before 0.3.26a3), for 2 values you need to use `Tense.probability2()` method \\
        instead, because inner code catches unexpected exception `ZeroDivisionError`. For version 0.3.25 this parameter accepts: 
        - integers
        - integer lists of size 1-2
        - integer tuples of size 1-2
        - integer deques of size 1-2
        - integer key-integer/`None`/`...` value dicts
        - integer sets and frozensets of size 1-2 both

        - `length` (Optional) - integer which has to be a denominator in probability fraction. Defaults to `-1`, what means this \\
        number is determined by `vf` passed values (simple integer is plus 1, dicts - plus value or 1 if it is `None` or ellipsis, \\
        sequence - 2nd item; if `None`, then plus 1). Since 0.3.26b3 put another restriction: length must be least than or equal \\
        `sys.maxsize`, which can be either equal 2\\*\\*31 - 1 (2,147,483,647) or 2\\*\\*63 - 1 (9,223,372,036,854,775,807)
        """
        # explanation:
        # a1 is final list, which will be used to return the integer
        # a2 is temporary list, which will store all single integers, without provided "frequency" value (item 2)
        a1, a2 = [[0] for _ in abroad(2)]
        self.clear(a1, a2)

        # c1 sums all instances of single numbers, and with provided "frequency" - plus it
        # c2 is substraction of length and c1
        # c3 has same purpose as c1, but counts only items without "frequency" (second) item
        # c4 is last variable, which is used as last from all of these - counts the modulo of c2 and c3 (but last one minus 1 as well)
        # c5 gets value from rearmost iteration 
        c1, c2, c3, c4, c5 = [0 for _ in abroad(5)]
        
        # 0.3.33: refraining from using string literals, since they will need to be manually changed
        # once parameter names are changed
        # note that 'return' keyword is reserved for return annotation with '->' operator
        _params = [k for k in self.probability.__annotations__ if k not in ("self", "return")]
        
        if not self.isInteger(length) and length != self.PROBABILITY_COMPUTE:
            error = TypeError("expected integer or constant '{}' in parameter '{}'".format(self.PROBABILITY_COMPUTE.name, _params[1]))
            raise error
        
        _length = -1 if length == self.PROBABILITY_COMPUTE else length
        
        # length checking before factual probability procedure
        if _length < -1:
            error = ValueError("expected integer value from -1 or above in parameter '{}'".format(_params[1]))
            raise error
        
        elif _length == 0:
            error = ZeroDivisionError("expected integer value from -1 or above in parameter '{}', but not equal zero".format(_params[1]))
            raise error
        
        # 0.3.26b3: cannot be greater than sys.maxsize
        # 0.3.33: update of this statement (added the 'or' keyword with expression as right operand)
        elif _length > sys.maxsize or (_length > sys.maxsize and reckon(vf) == 2 and not TenseOptions.disableProbability2LengthLimit):
            error = ValueError("integer value passed to parameter '{}' is too high, expected value below or equal {}".format(_params[1], sys.maxsize))
            raise error
        
        # START 0.3.26a3
        if reckon(vf) == 2:
            
            # alias to elements
            e1, e2 = (vf[0], vf[1])
            
            if self.isInteger(e1):
                
                if self.isInteger(e2):
                    
                    # 0.3.33: 'length = _length' instead of 'length = 2'
                    return self.probability2(e1, e2, length = _length)
                
                # 0.3.33: _ProbabilitySeq type alias to deputize (_ProbabilitySeqNoDict, dict) tuple
                # during type checking
                elif isinstance(e2, _ProbabilitySeq):
                    
                    if isinstance(e2, _ProbabilitySeqNoDict):
                    
                        if reckon(e2) in (1, 2):
                            
                            _tmp = e2[0] if reckon(e2) == 1 else (e2[0], e2[1])
                            
                            if self.isInteger(_tmp):
                                
                                # 0.3.33: 'length = _length' instead of 'length = 2'
                                return self.probability2(e1, _tmp, length = _length)
                            
                            # notable information: this duo of tuple items are, respectively:
                            # 'value' and 'frequency'
                            elif self.isTuple(_tmp):
                                
                                _v, _f = (_tmp[0], _tmp[1])
                                
                                if not self.isInteger(_v):
                                    error = TypeError("first item in an iterable is not an integer")
                                    raise error
                                
                                if not (self.isInteger(_f) or self.isEllipsis(_f) or self.isNone(_f)):
                                    error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                    raise error
                                
                                if self.isNone(_f) or self.isEllipsis(_f):
                                    return self.probability2(e1, _v, length = _length)
                                
                                elif _f < 1: # probability fraction cannot be negative
                                    error = ValueError("second item in an iterable is negative or equal zero")
                                    raise error
                                
                                return self.probability2(e1, _v, frequency = _f, length = _length)
                            
                            else:
                                error = RuntimeError("internal error")
                                raise error
                            
                        else:
                            error = IndexError("length of every iterable may have length 1-2 only")
                            raise error
                        
                    elif self.isDict(e2):
                        
                        if reckon(e2) != 1:
                            error = ValueError("expected one pair in a dictonary, received {}".format(reckon(e2)))
                            raise error
                        
                        # 0.3.33: removed loop, replacing it with inbuilt dictionary methods
                        _v, _f = (self.toList(e2.keys())[0], self.toList(e2.values())[0])
                        
                        if self.isNone(_f) or self.isEllipsis(_f):
                            return self.probability2(e1, _v, length = _length)
                        
                        elif _f < 1:
                            error = ValueError("value in a dictionary is negative or equal zero")
                            raise error
                            
                        return self.probability2(e1, _v, frequency = _f, length = _length)
                    
                    else:
                        error = TypeError("unsupported type in second item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                        raise error
            
            # 0.3.33: _ProbabilitySeq type alias to deputize (_ProbabilitySeqNoDict, dict) tuple
            # during type checking
            elif isinstance(e1, _ProbabilitySeq):
                
                if self.isInteger(e2):
                
                    if isinstance(e1, _ProbabilitySeqNoDict):
                        
                        if reckon(e1) in (1, 2):
                            
                            _tmp = e1[0] if reckon(e1) == 1 else (e1[0], e1[1])
                            
                            if self.isInteger(_tmp):
                                
                                # 0.3.33: 'length = _length' instead of 'length = 2'
                                # + 'e2' and '_tmp' are placed vice versa
                                return self.probability2(_tmp, e2, length = _length)
                            
                            elif self.isTuple(_tmp):
                                
                                _v, _f = (_tmp[0], _tmp[1])
                                
                                if not self.isInteger(_v):
                                    error = TypeError("first item in an iterable is not an integer")
                                    raise error
                                
                                if not (self.isInteger(_f) or self.isEllipsis(_f) or self.isNone(_f)):
                                    error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                    raise error
                                
                                if self.isNone(_f) or self.isEllipsis(_f):
                                    return self.probability2(_v, e2, length = _length)
                                
                                elif _f < 1: # probability fraction cannot be negative
                                    error = ValueError(f"second item in an iterable is negative or equal zero")
                                    raise error
                                
                                return self.probability2(_v, e2, frequency = _f, length = _length)
                            
                            else:
                                error = RuntimeError("internal error")
                                raise error
                            
                        else:
                            error = IndexError("length of every iterable may have length 1-2 only")
                            raise error
                    
                    elif self.isDict(e1):
                        
                        if reckon(e1) != 1:
                            error = ValueError("expected only one pair in a dictonary, received {}".format(reckon(e1)))
                            raise error
                        
                        # 0.3.33: removed loop, replacing it with inbuilt dictionary methods
                        _v, _f = (self.toList(e1.keys())[0], self.toList(e1.values())[0])
                        
                        if self.isNone(_f) or self.isEllipsis(_f):
                            
                            # 0.3.33: '_v' and 'e2' vice versa
                            return self.probability2(_v, e2, length = _length)
                        
                        elif _f < 1:
                            error = ValueError("value in a dictionary is negative or equal zero")
                            raise error
                        
                        # 0.3.33: '_v' and 'e2' vice versa
                        return self.probability2(_v, e2, frequency = _f, length = _length)
                
                # 0.3.33: _ProbabilitySeq type alias to deputize (_ProbabilitySeqNoDict, dict) tuple
                # during type checking
                elif isinstance(e2, _ProbabilitySeq):
                    
                    if isinstance(e1, _ProbabilitySeqNoDict):
                    
                        if isinstance(e2, _ProbabilitySeqNoDict):
                            
                            if reckon(e1) == 1:
                                
                                if reckon(e2) == 1:
                                    
                                    _1, _2 = (e1[0], e2[0])
                                    
                                    if not self.isInteger(_1) or not self.isInteger(_2):
                                        error = TypeError("first item in an iterable is not an integer")
                                        raise error
                                    
                                    return self.probability2(_1, _2, length = _length)
                                
                                elif reckon(e2) == 2:
                                    
                                    _1, _2_1, _2_2 = (e1[0], e2[0], e2[1])
                                    
                                    if not self.isInteger(_1) or not self.isInteger(_2_1):
                                        error = TypeError("first item in an iterable is not an integer")
                                        raise error
                                    
                                    if not self.isEllipsis(_2_2) and not self.isInteger(_2_2) and not self.isNone(_2_2):
                                        error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                        raise error
                                    
                                    if self.isNone(_2_2) or self.isEllipsis(_2_2):
                                        
                                        return self.probability2(_1, _2_1, length = _length)
                                    
                                    return self.probability2(_1, _2_1, frequency = _2_2, length = _length)
                                
                                else:
                                    error = IndexError("length of every iterable may have length 1-2 only")
                                    raise error
                            
                        elif reckon(e1) == 2:
                            
                            if reckon(e2) == 1:
                                
                                _1_1, _1_2, _2 = (e1[0], e1[1], e2[0])
                                
                                if not self.isInteger(_1_1) or not self.isInteger(_2):
                                    error = TypeError("first item in an iterable is not an integer")
                                    raise error
                                
                                if not self.isEllipsis(_1_2) and not self.isInteger(_1_2) and self.isNone(_1_2):
                                    error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                    raise error
                                
                                if self.isNone(_1_2) or self.isEllipsis(_1_2):
                                    
                                    return self.probability2(_1_1, _2, length = _length)
                                
                                return self.probability2(_1_1, _2, frequency = _length - _1_2, length = _length)
                            
                            elif reckon(e2) == 2:
                                
                                _1_1, _1_2, _2_1, _2_2 = (e1[0], e1[1], e2[0], e2[1])
                                
                                if not self.isInteger(_1_1) or not self.isInteger(_2_1):
                                    error = TypeError("first item in an iterable is not an integer")
                                    raise error
                                
                                if (
                                    not self.isEllipsis(_1_2) and not self.isInteger(_1_2) and self.isNone(_1_2)) or (
                                    not self.isEllipsis(_2_2) and not self.isInteger(_2_2) and self.isNone(_2_2)
                                ):
                                    error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                    raise error
                                
                                if self.isNone(_1_2) or self.isEllipsis(_1_2):
                                    
                                    if self.isNone(_2_2) or self.isEllipsis(_2_2):
                                        return self.probability2(_1_1, _2_1, length = _length)
                                    
                                    else:
                                        return self.probability2(_1_1, _2_1, frequency = _length - _2_2, length = _length)
                                    
                                elif self.isInteger(_1_2):
                                    
                                    if self.isNone(_2_2) or self.isEllipsis(_2_2):
                                        return self.probability2(_1_1, _2_1, frequency = _1_2, length = _length)
                                    
                                    else:
                                        return self.probability2(_1_1, _2_1, frequency = _1_2, length = _length if _length > _1_2 + _2_2 else _1_2 + _2_2)
                                    
                                else:
                                    error = RuntimeError("internal error")
                                    raise error
                                    
                            else:
                                error = IndexError("length of every iterable may have length 1-2 only")
                                raise error
                            
                        else:
                            error = IndexError("length of every iterable may have length 1-2 only")
                            raise error
                    
                    elif self.isDict(e1):
                        
                        if reckon(e1) != 1:
                            error = ValueError("expected only one pair in a dictonary, received {}".format(reckon(e1)))
                            raise error
                        
                        # 0.3.33: removed loop, replacing it with inbuilt dictionary methods
                        _v, _f = (self.toList(e1.keys())[0], self.toList(e1.values())[0])
                        
                        if isinstance(e2, _ProbabilitySeqNoDict):
                            
                            if self.isNone(_f) or self.isEllipsis(_f):
                                
                                if reckon(e2) in (1, 2):
                                    
                                    _tmp = e2[0] if reckon(e2) == 1 else (e2[0], e2[1])
                            
                                    if self.isInteger(_tmp):
                                        
                                        return self.probability2(_v, _tmp, length = _length)
                                    
                                    elif self.isTuple(_tmp):
                                        
                                        # avoid name collision
                                        _v2, _f2 = (_tmp[0], _tmp[1])
                                        
                                        if not self.isInteger(_v2):
                                            error = TypeError("first item in an iterable is not an integer")
                                            raise error
                                        
                                        if not (self.isInteger(_f2) or self.isEllipsis(_f2) or self.isNone(_f2)):
                                            error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                            raise error
                                        
                                        if self.isNone(_f2) or self.isEllipsis(_f2):
                                            return self.probability2(_v, _v2, length = _length)
                                        
                                        elif _f2 < 1:
                                            error = ValueError("second item in an iterable is negative or equal zero")
                                            raise error
                                        
                                        return self.probability2(_v, _v2, frequency = _length - _f2, length = _length)
                            
                                    else:
                                        error = RuntimeError("internal error")
                                        raise error
                                
                                else:
                                    error = IndexError("length of every iterable may have length 1-2 only")
                                    raise error
                            
                            elif _f < 1:
                                error = ValueError("value in a dictionary is negative or equal zero")
                                raise error
                            
                            if reckon(e2) in (1, 2):
                                    
                                _tmp = e2[0] if reckon(e2) == 1 else (e2[0], e2[1])
                        
                                if self.isInteger(_tmp):
                                    
                                    return self.probability2(_v, _tmp, length = _length)
                                
                                elif self.isTuple(_tmp):
                                    
                                    # avoid name collision
                                    _v2, _f2 = (_tmp[0], _tmp[1])
                                    
                                    if not self.isInteger(_v2):
                                        error = TypeError("first item in an iterable is not an integer")
                                        raise error
                                    
                                    if not (self.isInteger(_f2) or self.isEllipsis(_f2) or self.isNone(_f2)):
                                        error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                                        raise error
                                    
                                    if self.isNone(_f2) or self.isEllipsis(_f2):
                                        return self.probability2(_v, _v2, length = _length)
                                    
                                    elif _f2 < 1:
                                        error = ValueError("second item in an iterable is negative or equal zero")
                                        raise error
                                    
                                    return self.probability2(_v, _v2, frequency = _length - _f2, length = _length)
                                
                                else:
                                    error = RuntimeError("internal error")
                                    raise error
                            
                            else:
                                error = IndexError("length of every iterable may have length 1-2 only")
                                raise error
                                
                        elif self.isDict(e2):
                            
                            if reckon(e2) != 1:
                                error = ValueError("expected only one pair in a dictonary, received {}".format(reckon(e1)))
                                raise error
                            
                            _v2, _f2 = (self.toList(e2.keys())[0], self.toList(e2.values())[0])
                            
                            if self.isNone(_f) or self.isEllipsis(_f):
                                
                                if self.isNone(_f2) or self.isEllipsis(_f2):
                                    
                                    return self.probability2(_v, _v2, length = _length)
                                
                                elif _f2 < 1:
                                    error = ValueError("value in a dictionary is negative or equal zero")
                                    raise error
                                
                                return self.probability2(_v, _v2, frequency = _length - _f2, length = _length)
                            
                            elif _f < 1:
                                error = ValueError("value in a dictionary is negative or equal zero")
                                raise error
                            
                            return self.probability2(_v, _v2, frequency = _f, length = _length if _length > _f + _f2 else _f + _f2)
                            
                            
                        else:
                            error = TypeError("unsupported type in second item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                            raise error
                    else:
                        error = TypeError("unsupported type in first item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                        raise error
                else:
                    error = TypeError("unsupported type in first item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                    raise error
            else:
                error = TypeError("unsupported type in first item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                raise error         
                    
        # END 0.3.26a3
        
        elif reckon(vf) < 2:
            
            error = _tc.MissingValueError("expected at least 2 items in variable parameter '{}', received {}".format(_params[0], reckon(vf)))
            raise error
        
        # reading all items provided
        for e in vf:
            
            # value is an integer (that means it cannot have "frequency"),
            # which is "value" parameter equivalent
            if isinstance(e, int):
                a1.append(e)
                a2.append(e)
                c1 += 1
                c3 += 1
                
            elif isinstance(e, _ProbabilitySeqNoDict):
                # we have only one item, and that item is "value"
                # 0.3.25 (additional statement for tuple and overall)
                if reckon(e) == 1:
                    
                    if not self.isInteger(e[0]):
                        error = TypeError("every iterable may have integer as first item")
                        raise error
                    
                    a1.append(e[0])
                    a2.append(e[0])
                    c1 += 1
                    c3 += 1
                    
                # those are, respectively, "value" and "frequency"
                elif reckon(e) == 2:
                    
                    if not self.isInteger(e[0]):
                        error = TypeError("every iterable may have integer as first item")
                        raise error
                    
                    if not (self.isInteger(e[1]) or self.isEllipsis(e[1]) or self.isNone(e[1])):
                        error = TypeError("second item in an iterable is neither an integer, 'None', nor an ellipsis")
                        raise error
                    
                    if self.isEllipsis(e[1]) or self.isNone(e[1]):
                        a1.append(e[0])
                        a2.append(e[0])
                        c1 += 1
                        c3 += 1
                    
                    elif e[1] < 1:
                        error = ValueError("second item in an iterable is a negative integer")
                        raise error
                    
                    for _ in abroad(e[1]): a1.append(e[0])
                    c1 += int(e[1])
                    
                # if thought that the length is third item, that is wrong presupposition
                else:
                    error = IndexError("length of every iterable may have length 1-2 only")
                    raise error
                
            # 0.3.24 (dict support)
            elif self.isDict(e):
                
                if reckon(e) == 0:
                   error = ValueError("Expected at least one pair in every dictonary, received {}".format(reckon(e)))
                   raise error
               
                for f in e:
                    
                    if not self.isInteger(f):
                        error = KeyError(f"One of keys in dictionaries is not an integer. Ensure every key is of type 'int'. Error thrown by item: '{f}'")
                        raise error
                    
                    if not (self.isInteger(e[f]) or self.isEllipsis(e[f]) or self.isNone(e[f])):
                        error = ValueError(f"One of values in dictionaries is neither an integer, 'None', nor an ellipsis. Ensure every values satisfies this requirement. Error thrown by item: '{f}'")
                        raise error
                    
                    if e[f] < 1:
                        error = ValueError(f"One of values in dictionaries is negative integer or equal zero. Ensure every value is positive integer. Error thrown by item: '{e[f]}'")
                        raise error
                    
                    elif self.isEllipsis(e[f]) or self.isNone(e[f]):
                        a1.append(f)
                        a2.append(f)
                        c1 += 1
                        c3 += 1
                        
                    else:
                        for _ in abroad(e[f]): a1.append(f)
                        c1 += 1
                        
            # incorrect type defined
            else:
                error = TypeError("unsupported type in an item of variable parameter '{}', expected an integer, iterable with 1-2 items (first item being integer), or dictionary with one pair (key being integer)".format(_params[0]))
                raise error   
            
        # length minus times when single integers are provided is needed
        # to continue extended probability
        if length == self.PROBABILITY_COMPUTE: c2 = c1
        else: c2 = _length - c1
        
        # hint: if not that minus one, last item will be also included
        # and we want the modulo just for the last item
        if c3 > 1: c3 -= 1
        
        # that instruction shouldn't be complicated
        # also, it is impossible to get into here, since
        # most values do not have "frequency" (aka 2nd item)
        if c2 != c1 and c2 > _length:
            tmp = [0]
            tmp.clear()
            for i in abroad(_length): tmp.append(a1[i])
            a1 = tmp
            
        # look in here: used is abroad() method, but before the valid
        # loop, all items of temporary variable will become positive
        elif c2 == c1 or (c2 != c1 and c2 < _length):
            
            for i in abroad(c2):
                
                # there we are looking for the highest number, which will
                # be divisible by number of integers passed to "vf" parameter
                if i % c3 == 0:
                    c5 = i
                    break
            # this loop is nested as we use to repeat all items from a2 list
            for i in abroad(a2):
                for _ in abroad(c5 / c3):
                    a1.append(a2[i])
                    
            # modulo will be used merely there
            c4 = c2 % c3
            
            # that one will be always done (more likely)
            # only indicated whether result in more than zero
            if c4 > 0:
                for _ in abroad(c4): a1.append(a2[reckon(a2) - 1])

        # code with following 'if False' below: as a general warning, you would use Tense.architecture()
        # to find your system's architecture, because it determines about value of sys.maxsize, which is
        # max size for all sequences not entirely sure, if creating 2d sequences to increase 'length'
        # parameter value is a good idea. scrap code below is projected
        if False:
            a3 = [[0]]
            a3.clear()
            _2d_i = 0
            while _2d_i < sys.maxsize:
                for _ in abroad(sys.maxsize):
                    a3[_2d_i].append(self.pick(a1))
                _2d_i += 1
            return self.pick(self.pick(a3))
        
        return self.pick(a1) # this will be returned
    
    @classmethod
    def until(self, desiredString: _SizeableItemGetter[str], /, message: _opt[str] = None, caseInsensitive: bool = True):
        """
        \\@since 0.3.25
        ```ts
        "class method" in class Tense
        ```
        Console method, which will repeat the program until user won't \\
        write correct string. Case is insensitive, may be configured via \\
        optional parameter `caseInsensitive`, which by default has \\
        value `True`. Returned is reference to this class.
        """
        s = ""
        c = False
        _d = desiredString
        _d = list(_d) if isinstance(_d, (set, frozenset)) else _d
        
        while c:
            s = input(message if message is not None and message != "" else "")
            c = s.lower() != _d.lower() if self.isString(_d) else s.lower() not in (_s.lower() for _s in _d)
            if not caseInsensitive:
                c = s != _d if self.isString(_d) else s not in _d
        return self
    
    @classmethod
    def sleep(self, seconds: float, /):
        """
        \\@since 0.3.25
        ```ts
        "class method" in class Tense
        ```
        Define an execution delay, which can be a floating-point number \\
        with 2 fractional digits. Returned is reference to this class.
        """
        _time.sleep(seconds)
        return self
    
    @classmethod
    def repeat(self, value: _T, times: _opt[int] = None, /): # 0.3.31: default value 'None'
        """
        \\@since 0.3.25 `itertools.repeat`
        ```ts
        "class method" in class Tense
        ```
        Returns list with `value` repeated `times` times. \\
        Equals `itertools.repeat()` (0.3.31).
        """
        
        return _itertools.repeat(value, times) if not self.isNone(times) else _itertools.repeat(value)
    
    
    @classmethod
    def cycle(self, i: _tc.Iterable[_T], /):
        """
        \\@since 0.3.34 `itertools.cycle`
        """
        return _itertools.cycle(i)
    
    
    count = _itertools.count # do not initialize
    """\\@since 0.3.34"""
    
    
    if TenseOptions.enableFurryDeclarations: # since 0.3.34
        
        @classmethod
        def __local_owoify_template(self, string: str, /) -> str:
            "\\@since 0.3.26b1"
            s: str = string
            s = _re.sub(r"\s{2}", " UrU ", s, flags = _re.M)
            s = _re.sub(r"XD", "UrU", s, flags = _re.M | _re.I)
            s = _re.sub(r":D", "UrU", s, flags = _re.M)
            s = _re.sub(r"lenny face", "OwO", s, flags = _re.M | _re.I)
            s = _re.sub(r":O", "OwO", s, flags = _re.M | _re.I)
            s = _re.sub(r":\)", ":3", s, flags = _re.M)
            s = _re.sub(r"\?", " uwu?", s, flags = _re.M) # ? is a metachar
            s = _re.sub(r"!", " owo!", s, flags = _re.M)
            s = _re.sub(r"; ", "~ ", s, flags = _re.M)
            s = _re.sub(r", ", "~ ", s, flags = _re.M)
            s = _re.sub(r"you are", "chu is", s, flags = _re.M)
            s = _re.sub(r"You are", "chu is".capitalize(), s, flags = _re.M)
            s = _re.sub(r"You Are", "chu is".title(), s, flags = _re.M)
            s = _re.sub(r"YOU ARE", "chu is".upper(), s, flags = _re.M)
            s = _re.sub(r"wat's this", "OwO what's this", s, flags = _re.M)
            s = _re.sub(r"Wat's [Tt]his", "OwO What's this", s, flags = _re.M)
            s = _re.sub(r"WAT'S THIS", "OwO what's this".upper(), s, flags = _re.M)
            s = _re.sub(r"old person", "greymuzzle", s, flags = _re.M)
            s = _re.sub(r"Old [Pp]erson", "greymuzzle".capitalize(), s, flags = _re.M)
            s = _re.sub(r"OLD PERSON", "greymuzzle".upper(), s, flags = _re.M)
            s = _re.sub(r"forgive me father, I have sinned", "sowwy daddy~ I have been naughty", s, flags = _re.M)
            s = _re.sub(r"Forgive me father, I have sinned", "sowwy daddy~ I have been naughty".capitalize(), s, flags = _re.M)
            s = _re.sub(r"FORGIVE ME FATHER, I HAVE SINNED", "sowwy daddy~ I have been naughty".upper(), s, flags = _re.M)
            s = _re.sub(r"your ", "ur ", s, flags = _re.M)
            s = _re.sub(r"Your ", "Ur ", s, flags = _re.M)
            s = _re.sub(r"YOUR ", "UR ", s, flags = _re.M)
            s = _re.sub(r" your", " ur", s, flags = _re.M)
            s = _re.sub(r" Your", " Ur", s, flags = _re.M)
            s = _re.sub(r" YOUR", " UR", s, flags = _re.M)
            s = _re.sub(r"(^your)| your", "ur", s, flags = _re.M)
            s = _re.sub(r"(^Your)| Your", "Ur", s, flags = _re.M)
            s = _re.sub(r"(^YOUR)| YOUR", "UR", s, flags = _re.M)
            s = _re.sub(r"you", "chu", s, flags = _re.M)
            s = _re.sub(r"You", "Chu", s, flags = _re.M)
            s = _re.sub(r"YOU", "CHU", s, flags = _re.M)
            s = _re.sub(r"with ", "wif ", s, flags = _re.M)
            s = _re.sub(r"With ", "Wif ", s, flags = _re.M)
            s = _re.sub(r"wITH ", "wIF ", s, flags = _re.M)
            s = _re.sub(r"what", "wat", s, flags = _re.M)
            s = _re.sub(r"What", "Wat", s, flags = _re.M)
            s = _re.sub(r"WHAT", "WAT", s, flags = _re.M)
            s = _re.sub(r"toe", "toe bean", s, flags = _re.M)
            s = _re.sub(r"Toe", "Toe Bean", s, flags = _re.M)
            s = _re.sub(r"TOE", "TOE BEAN", s, flags = _re.M)
            s = _re.sub(r"this", "dis", s, flags = _re.M)
            s = _re.sub(r"This", "Dis", s, flags = _re.M)
            s = _re.sub(r"THIS", "DIS", s, flags = _re.M)
            s = _re.sub(r"(?!hell\w+)hell", "hecc", s, flags = _re.M)
            s = _re.sub(r"(?!Hell\w+)Hell", "Hecc", s, flags = _re.M)
            s = _re.sub(r"(?!HELL\w+)HELL", "HECC", s, flags = _re.M)
            s = _re.sub(r"the ", "teh ", s, flags = _re.M)
            s = _re.sub(r"^the$", "teh", s, flags = _re.M)
            s = _re.sub(r"The ", "Teh ", s, flags = _re.M)
            s = _re.sub(r"^The$", "Teh", s, flags = _re.M)
            s = _re.sub(r"THE ", "TEH ", s, flags = _re.M)
            s = _re.sub(r"^THE$", "TEH", s, flags = _re.M)
            s = _re.sub(r"tare", "tail", s, flags = _re.M)
            s = _re.sub(r"Tare", "Tail", s, flags = _re.M)
            s = _re.sub(r"TARE", "TAIL", s, flags = _re.M)
            s = _re.sub(r"straight", "gay", s, flags = _re.M)
            s = _re.sub(r"Straight", "Gay", s, flags = _re.M)
            s = _re.sub(r"STRAIGHT", "GAY", s, flags = _re.M)
            s = _re.sub(r"source", "sauce", s, flags = _re.M)
            s = _re.sub(r"Source", "Sauce", s, flags = _re.M)
            s = _re.sub(r"SOURCE", "SAUCE", s, flags = _re.M)
            s = _re.sub(r"(?!slut\w+)slut", "fox", s, flags = _re.M)
            s = _re.sub(r"(?!Slut\w+)Slut", "Fox", s, flags = _re.M)
            s = _re.sub(r"(?!SLUT\w+)SLUT", "FOX", s, flags = _re.M)
            s = _re.sub(r"shout", "awoo", s, flags = _re.M)
            s = _re.sub(r"Shout", "Awoo", s, flags = _re.M)
            s = _re.sub(r"SHOUT", "AWOO", s, flags = _re.M)
            s = _re.sub(r"roar", "rawr", s, flags = _re.M)
            s = _re.sub(r"Roar", "Rawr", s, flags = _re.M)
            s = _re.sub(r"ROAR", "RAWR", s, flags = _re.M)
            s = _re.sub(r"pawlice department", "paw patrol", s, flags = _re.M)
            s = _re.sub(r"Paw[Ll]ice [Dd]epartment", "Paw Patrol", s, flags = _re.M)
            s = _re.sub(r"PAWLICE DEPARTMENT", "PAW PATROL", s, flags = _re.M)
            s = _re.sub(r"police", "pawlice", s, flags = _re.M)
            s = _re.sub(r"Police", "Pawlice", s, flags = _re.M)
            s = _re.sub(r"POLICE", "PAWLICE", s, flags = _re.M)
            s = _re.sub(r"pervert", "furvert", s, flags = _re.M)
            s = _re.sub(r"Pervert", "Furvert", s, flags = _re.M)
            s = _re.sub(r"PERVERT", "FURVERT", s, flags = _re.M)
            s = _re.sub(r"persona", "fursona", s, flags = _re.M)
            s = _re.sub(r"Persona", "Fursona", s, flags = _re.M)
            s = _re.sub(r"PERSONA", "FURSONA", s, flags = _re.M)
            s = _re.sub(r"perfect", "purrfect", s, flags = _re.M)
            s = _re.sub(r"Perfect", "Purrfect", s, flags = _re.M)
            s = _re.sub(r"PERFECT", "PURRFECT", s, flags = _re.M)
            s = _re.sub(r"(?!not\w+)not", "nawt", s, flags = _re.M)
            s = _re.sub(r"(?!Not\w+)Not", "Nawt", s, flags = _re.M)
            s = _re.sub(r"(?!NOT\w+)NOT", "NAWT", s, flags = _re.M)
            s = _re.sub(r"naughty", "nawt", s, flags = _re.M)
            s = _re.sub(r"Naughty", "Nawt", s, flags = _re.M)
            s = _re.sub(r"NAUGHTY", "NAWT", s, flags = _re.M)
            s = _re.sub(r"name", "nyame", s, flags = _re.M)
            s = _re.sub(r"Name", "Nyame", s, flags = _re.M)
            s = _re.sub(r"NAME", "NYAME", s, flags = _re.M)
            s = _re.sub(r"mouth", "maw", s, flags = _re.M)
            s = _re.sub(r"Mouth", "Maw", s, flags = _re.M)
            s = _re.sub(r"MOUTH", "MAW", s, flags = _re.M)
            s = _re.sub(r"love", "luv", s, flags = _re.M)
            s = _re.sub(r"Love", "Luv", s, flags = _re.M)
            s = _re.sub(r"LOVE", "LUV", s, flags = _re.M)
            s = _re.sub(r"lol", "waw", s, flags = _re.M)
            s = _re.sub(r"Lol", "Waw", s, flags = _re.M)
            s = _re.sub(r"LOL", "WAW", s, flags = _re.M)
            s = _re.sub(r"lmao", "hehe~", s, flags = _re.M)
            s = _re.sub(r"Lmao", "Hehe~", s, flags = _re.M)
            s = _re.sub(r"LMAO", "HEHE~", s, flags = _re.M)
            s = _re.sub(r"kiss", "lick", s, flags = _re.M)
            s = _re.sub(r"Kiss", "Lick", s, flags = _re.M)
            s = _re.sub(r"KISS", "LICK", s, flags = _re.M)
            s = _re.sub(r"lmao", "hehe~", s, flags = _re.M)
            s = _re.sub(r"Lmao", "Hehe~", s, flags = _re.M)
            s = _re.sub(r"LMAO", "HEHE~", s, flags = _re.M)
            s = _re.sub(r"hyena", "yeen", s, flags = _re.M)
            s = _re.sub(r"Hyena", "Yeen", s, flags = _re.M)
            s = _re.sub(r"HYENA", "YEEN", s, flags = _re.M)
            s = _re.sub(r"^hi$", "hai", s, flags = _re.M)
            s = _re.sub(r" hi ", " hai~ ", s, flags = _re.M)
            s = _re.sub(r"hi(,| )", "hai~ ", s, flags = _re.M)
            s = _re.sub(r"hi!", "hai!", s, flags = _re.M)
            s = _re.sub(r"hi\?", "hai?", s, flags = _re.M)
            s = _re.sub(r"^Hi$", "Hai", s, flags = _re.M)
            s = _re.sub(r" Hi ", " Hai~ ", s, flags = _re.M)
            s = _re.sub(r"Hi(,| )", "Hai~ ", s, flags = _re.M)
            s = _re.sub(r"Hi!", "Hai!", s, flags = _re.M)
            s = _re.sub(r"Hi\?", "Hai?", s, flags = _re.M)
            s = _re.sub(r"^HI$", "HAI", s, flags = _re.M)
            s = _re.sub(r" HI ", " HAI~ ", s, flags = _re.M)
            s = _re.sub(r"HI(,| )", "HAI~ ", s, flags = _re.M)
            s = _re.sub(r"HI!", "HAI!", s, flags = _re.M)
            s = _re.sub(r"HI\?", "HAI?", s, flags = _re.M)
            s = _re.sub(r"(?!handy)hand", "paw", s, flags = _re.M)
            s = _re.sub(r"(?!Handy)Hand", "Paw", s, flags = _re.M)
            s = _re.sub(r"(?!HANDY)HAND", "PAW", s, flags = _re.M)
            s = _re.sub(r"handy", "pawi", s, flags = _re.M)
            s = _re.sub(r"Handy", "Pawi", s, flags = _re.M)
            s = _re.sub(r"HANDY", "PAWI", s, flags = _re.M)
            s = _re.sub(r"for", "fur", s, flags = _re.M)
            s = _re.sub(r"For", "Fur", s, flags = _re.M)
            s = _re.sub(r"FOR", "FUR", s, flags = _re.M)
            s = _re.sub(r"foot", "footpaw", s, flags = _re.M)
            s = _re.sub(r"Foot", "Footpaw", s, flags = _re.M)
            s = _re.sub(r"FOOT", "FOOTPAW", s, flags = _re.M)
            s = _re.sub(r"father", "daddy", s, flags = _re.M)
            s = _re.sub(r"Father", "Daddy", s, flags = _re.M)
            s = _re.sub(r"FATHER", "DADDY", s, flags = _re.M)
            s = _re.sub(r"fuck", "fluff", s, flags = _re.M)
            s = _re.sub(r"Fuck", "Fluff", s, flags = _re.M)
            s = _re.sub(r"FUCK", "FLUFF", s, flags = _re.M)
            s = _re.sub(r"dragon", "derg", s, flags = _re.M)
            s = _re.sub(r"Dragon", "Derg", s, flags = _re.M)
            s = _re.sub(r"DRAGON", "DERG", s, flags = _re.M)
            s = _re.sub(r"(?!doggy)dog", "good boi", s, flags = _re.M)
            s = _re.sub(r"(?!Doggy)Dog", "Good boi", s, flags = _re.M)
            s = _re.sub(r"(?!DOGGY)DOG", "GOOD BOI", s, flags = _re.M)
            s = _re.sub(r"disease", "pathOwOgen", s, flags = _re.M)
            s = _re.sub(r"Disease", "PathOwOgen", s, flags = _re.M)
            s = _re.sub(r"DISEASE", "PATHOWOGEN", s, flags = _re.M)
            s = _re.sub(r"cyborg|robot|computer", "protogen", s, flags = _re.M)
            s = _re.sub(r"Cyborg|Robot|Computer", "Protogen", s, flags = _re.M)
            s = _re.sub(r"CYBORG|ROBOT|COMPUTER", "PROTOGEN", s, flags = _re.M)
            s = _re.sub(r"(?!children)child", "cub", s, flags = _re.M)
            s = _re.sub(r"(?!Children)Child", "Cub", s, flags = _re.M)
            s = _re.sub(r"(?!CHILDREN)CHILD", "CUB", s, flags = _re.M)
            s = _re.sub(r"(?!cheese[ds])cheese", "sergal", s, flags = _re.M)
            s = _re.sub(r"(?!Cheese[ds])Cheese", "Sergal", s, flags = _re.M)
            s = _re.sub(r"(?!CHEESE[DS])CHEESE", "SERGAL", s, flags = _re.M)
            s = _re.sub(r"celebrity", "popufur", s, flags = _re.M)
            s = _re.sub(r"Celebrity", "Popufur", s, flags = _re.M)
            s = _re.sub(r"CELEBRITY", "POPUFUR", s, flags = _re.M)
            s = _re.sub(r"bye", "bai", s, flags = _re.M)
            s = _re.sub(r"Bye", "Bai", s, flags = _re.M)
            s = _re.sub(r"BYE", "BAI", s, flags = _re.M)
            s = _re.sub(r"butthole", "tailhole", s, flags = _re.M)
            s = _re.sub(r"Butthole", "Tailhole", s, flags = _re.M)
            s = _re.sub(r"BUTTHOLE", "TAILHOLE", s, flags = _re.M)
            s = _re.sub(r"bulge", "bulgy-wulgy", s, flags = _re.M)
            s = _re.sub(r"Bulge", "Bulgy-wulgy", s, flags = _re.M)
            s = _re.sub(r"BULGE", "BULGY-WULGY", s, flags = _re.M)
            s = _re.sub(r"bite", "nom", s, flags = _re.M)
            s = _re.sub(r"Bite", "Nom", s, flags = _re.M)
            s = _re.sub(r"BITE", "NOM", s, flags = _re.M)
            s = _re.sub(r"awful", "pawful", s, flags = _re.M)
            s = _re.sub(r"Awful", "Pawful", s, flags = _re.M)
            s = _re.sub(r"AWFUL", "PAWFUL", s, flags = _re.M)
            s = _re.sub(r"awesome", "pawsome", s, flags = _re.M)
            s = _re.sub(r"Awesome", "Pawsome", s, flags = _re.M)
            s = _re.sub(r"AWESOME", "PAWSOME", s, flags = _re.M)
            s = _re.sub(r"(?!ahh(h)+)ahh", "murr", s, flags = _re.M)
            s = _re.sub(r"(?!Ahh[Hh]+)Ahh", "Murr", s, flags = _re.M)
            s = _re.sub(r"(?!AHH(H)+)AHH", "MURR", s, flags = _re.M)
            s = _re.sub(r"(?![Gg]reymuzzle|[Tt]ail(hole)?|[Pp]aw [Pp]atrol|[Pp]awlice|luv|lick|[Ff]luff|[Ss]ergal|[Pp]awful)l", "w", s, flags = _re.M)
            s = _re.sub(r"(?!GREYMUZZLE|TAIL(HOLE)?|PAW PATROL|PAWLICE|L(uv|UV)|L(ick|ICK)|FLUFF|SERGAL|PAWFUL)L", "W", s, flags = _re.M)
            s = _re.sub(r"(?![Gg]reymuzzle|ur|[Rr]awr|[Ff]ur(sona|vert)?|[Pp]urrfect|[Vv]ore|[Dd]erg|[Pp]rotogen|[Ss]ergal|[Pp]opufur|[Mm]urr)r", "w", s, flags = _re.M)
            s = _re.sub(r"(?!GREYMUZZLE|UR|RAWR|FUR(SONA|VERT)?|PURRFECT|VORE|DERG|PROTOGEN|SERGAL|POPUFUR|MURR)R", "W", s, flags = _re.M)
            # above: 0.3.26a3, below: 0.3.26b1
            s = _re.sub(r"gweymuzzwe", "greymuzzle", s, flags = _re.M)
            s = _re.sub(r"Gweymuzzwe", "Greymuzzle", s, flags = _re.M)
            s = _re.sub(r"GWEYMUZZWE", "GREYMUZZLE", s, flags = _re.M)
            s = _re.sub(r"taiwhowe", "tailhole", s, flags = _re.M)
            s = _re.sub(r"Taiwhowe", "Tailhole", s, flags = _re.M)
            s = _re.sub(r"TAIWHOWE", "TAILHOLE", s, flags = _re.M)
            s = _re.sub(r"paw patwow", "paw patrol", s, flags = _re.M)
            s = _re.sub(r"Paw Patwow", "Paw Patrol", s, flags = _re.M)
            s = _re.sub(r"PAW PATWOW", "PAW PATROL", s, flags = _re.M)
            s = _re.sub(r"pawwice", "pawlice", s, flags = _re.M)
            s = _re.sub(r"Pawwice", "Pawlice", s, flags = _re.M)
            s = _re.sub(r"PAWWICE", "PAWLICE", s, flags = _re.M)
            s = _re.sub(r"wuv", "luv", s, flags = _re.M)
            s = _re.sub(r"Wuv", "Luv", s, flags = _re.M)
            s = _re.sub(r"WUV", "LUV", s, flags = _re.M)
            s = _re.sub(r"wick", "lick", s, flags = _re.M)
            s = _re.sub(r"Wick", "Lick", s, flags = _re.M)
            s = _re.sub(r"WICK", "LICK", s, flags = _re.M)
            s = _re.sub(r"fwuff", "fluff", s, flags = _re.M)
            s = _re.sub(r"Fwuff", "Fluff", s, flags = _re.M)
            s = _re.sub(r"FWUFF", "FLUFF", s, flags = _re.M)
            s = _re.sub(r"sewgaw", "sergal", s, flags = _re.M)
            s = _re.sub(r"Sewgaw", "Sergal", s, flags = _re.M)
            s = _re.sub(r"SEWGAW", "SERGAL", s, flags = _re.M)
            s = _re.sub(r"pawfuw", "pawful", s, flags = _re.M)
            s = _re.sub(r"Pawfuw", "Pawful", s, flags = _re.M)
            s = _re.sub(r"PAWFUW", "PAWFUL", s, flags = _re.M)
            s = _re.sub(r"(?!uwu)uw", "ur", s, flags = _re.M)
            s = _re.sub(r"(?!Uwu)Uw", "Ur", s, flags = _re.M)
            s = _re.sub(r"(?!UWU)UW", "UR", s, flags = _re.M)
            s = _re.sub(r"waww", "rawr", s, flags = _re.M)
            s = _re.sub(r"Waww", "Rawr", s, flags = _re.M)
            s = _re.sub(r"WAWW", "RAWR", s, flags = _re.M)
            s = _re.sub(r"fuw", "fur", s, flags = _re.M)
            s = _re.sub(r"Fuw", "Fur", s, flags = _re.M)
            s = _re.sub(r"FUW", "FUR", s, flags = _re.M)
            s = _re.sub(r"furvewt", "furvert", s, flags = _re.M)
            s = _re.sub(r"Furvewt", "Furvert", s, flags = _re.M)
            s = _re.sub(r"FURVEWT", "FURVERT", s, flags = _re.M)
            s = _re.sub(r"puwwfect", "purrfect", s, flags = _re.M)
            s = _re.sub(r"Puwwfect", "Purrfect", s, flags = _re.M)
            s = _re.sub(r"PUWWFECT", "PURRFECT", s, flags = _re.M)
            s = _re.sub(r"vowe", "vore", s, flags = _re.M)
            s = _re.sub(r"Vowe", "Vore", s, flags = _re.M)
            s = _re.sub(r"VOWE", "VORE", s, flags = _re.M)
            s = _re.sub(r"dewg", "derg", s, flags = _re.M)
            s = _re.sub(r"Dewg", "Derg", s, flags = _re.M)
            s = _re.sub(r"DEWG", "DERG", s, flags = _re.M)
            s = _re.sub(r"pwotogen", "protogen", s, flags = _re.M)
            s = _re.sub(r"Pwotogen", "Protogen", s, flags = _re.M)
            s = _re.sub(r"PWOTOGEN", "PROTOGEN", s, flags = _re.M)
            s = _re.sub(r"popufuw", "popufur", s, flags = _re.M)
            s = _re.sub(r"Popufuw", "Popufur", s, flags = _re.M)
            s = _re.sub(r"POPUFUW", "POPUFUR", s, flags = _re.M)
            s = _re.sub(r"muww", "murr", s, flags = _re.M)
            s = _re.sub(r"Muww", "Murr", s, flags = _re.M)
            s = _re.sub(r"MUWW", "MURR", s, flags = _re.M)
            # end 0.3.26b1; start 0.3.26rc2
            s = _re.sub(r"furwy", "fuwwy", s, flags = _re.M)
            s = _re.sub(r"Furwy", "Fuwwy", s, flags = _re.M)
            s = _re.sub(r"FURWY", "FUWWY", s, flags = _re.M)
            s = _re.sub(r"UrU", "UwU", s, flags = _re.M)
            s = _re.sub(r"Uru", "Uwu", s, flags = _re.M)
            s = _re.sub(r"uru", "uwu", s, flags = _re.M)
            s = _re.sub(r"URU", "UWU", s, flags = _re.M)
            s = _re.sub(r"femboy", "femboi", s, flags = _re.M)
            s = _re.sub(r"Femboy", "Femboi", s, flags = _re.M)
            s = _re.sub(r"FEMBOY", "FEMBOI", s, flags = _re.M)
            s = _re.sub(r":<", "x3", s, flags = _re.M)
            # end 0.3.26rc2; start 0.3.26
            s = _re.sub(r"ding", "beep", s, flags = _re.M)
            s = _re.sub(r"Ding", "Beep", s, flags = _re.M)
            s = _re.sub(r"DING", "BEEP", s, flags = _re.M)
            s = _re.sub(r"shourd", "shouwd", s, flags = _re.M)
            s = _re.sub(r"Shourd", "Shouwd", s, flags = _re.M)
            s = _re.sub(r"SHOURD", "SHOUWD", s, flags = _re.M)
            s = _re.sub(r"course", "couwse", s, flags = _re.M)
            s = _re.sub(r"Course", "Couwse", s, flags = _re.M)
            s = _re.sub(r"COURSE", "COUWSE", s, flags = _re.M)
            return s
        
        @classmethod
        def owoify(self, s: str, /):
            """
            \\@since 0.3.9 \\
            \\@lifetime ≥ 0.3.9; < 0.3.24; ≥ 0.3.25
            ```
            "class method" in class Tense
            ```
            Joke method translating a string to furry equivalent. \\
            Basing on https://lingojam.com/FurryTalk. Several words \\
            aren't included normally (0.3.26a3, 0.3.26b1, 0.3.26rc2, \\
            0.3.26), still, most are, several have different translations
            """
            return self.__local_owoify_template(s)
        
        @classmethod
        def uwuify(self, s: str, /):
            """
            \\@since 0.3.27b2 \\
            \\@lifetime ≥ 0.3.27b2
            ```
            "class method" in class Tense
            ```
            Alias to `Tense.owoify()`
            """
            return self.__local_owoify_template(s)
    
    @classmethod
    def aeify(self, s: str, /):
        """
        \\@since 0.3.9 \\
        \\@lifetime ≥ 0.3.9; < 0.3.24; ≥ 0.3.26a4
        ```
        "class method" in class Tense
        ```
        Joke method which converts every a and e into \u00E6. Ensure your \\
        compiler reads characters from ISO/IEC 8859-1 encoding, because \\
        without it you might meet question marks instead
        """
        _s = s
        _s = _re.sub(r"[ae]", "\u00E6", _s, flags = _re.M)
        _s = _re.sub(r"[AE]", "\u00C6", _s, flags = _re.M)
        return _s
    
    @classmethod
    def oeify(self, s: str, /):
        """
        \\@since 0.3.9 \\
        \\@lifetime ≥ 0.3.9; < 0.3.24; ≥ 0.3.26a4
        ```
        "class method" in class Tense
        ```
        Joke method which converts every o and e into \u0153. Ensure your \\
        compiler reads characters from ISO/IEC 8859-1 encoding, because \\
        without it you might meet question marks instead
        """
        _s = s
        _s = _re.sub(r"[oe]", "\u0153", _s, flags = _re.M)
        _s = _re.sub(r"[OE]", "\u0152", _s, flags = _re.M)
        return _s
    __all__ = [n for n in locals() if n[:1] != "_"]
    "\\@since 0.3.26rc2"
    __dir__ = [n for n in locals() if n[:1] != "_"]
    "\\@since 0.3.26rc2"

tense = Tense
TS = Tense()
"\\@since 0.3.27a5. Instance of `Tense` to use for `>>` and `<<` operators especially"


class RGB(_util.Final):
    """
    \\@since 0.3.28
    
    ```
    class RGB: ...
    from tense import RGB
    ```
    
    Auxiliary class for `Color` class. Represents red-green-blue color representation.
    """
    def __init__(self, red = 0, green = 0, blue = 0, /):
        
        _parameters = {
            "red": red,
            "green": green,
            "blue": blue
        }
        
        for key in _parameters:
            
            if not isinstance(_parameters[key], int) or (isinstance(_parameters[key], int) and _parameters[key] not in abroad(0x100)):
                error = TypeError("expected a non-negative integer in parameter '{}' in range 0-255".format(key))
                raise error
            
        self.__rgb = (red, green, blue)
        
    def __str__(self):
        """
        \\@since 0.3.28
        """
        return type(self).__name__ + "({})".format(", ".join([str(e) for e in self.__rgb]))
        
    def __repr__(self):
        """
        \\@since 0.3.28
        """
        return "<{} object: {}>".format(__module__ + "." + type(self).__name__, self.__str__())
    
    def __hex__(self):
        """
        \\@since 0.3.28
        
        Provides conversion to hexadecimal format
        """
        _r = hex(self.__rgb[0])[2:] if self.__rgb[0] >= 0x10 else "0" + hex(self.__rgb[0])[2:]
        _g = hex(self.__rgb[1])[2:] if self.__rgb[1] >= 0x10 else "0" + hex(self.__rgb[1])[2:]
        _b = hex(self.__rgb[2])[2:] if self.__rgb[2] >= 0x10 else "0" + hex(self.__rgb[2])[2:]
        return "0x" + _r + _g + _b
    
    def __int__(self):
        """
        \\@since 0.3.28
        
        Converts RGB tuple into its corresponding integer representation
        """
        return int(self.__hex__()[2:], base = 16)
    
    # little deviation from type hinting in methods below
    # read document strings to figure it out
    def __lt__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) < int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() < other.__int__() if isinstance(other, type(self)) else False
    
    def __gt__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) > int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() > other.__int__() if isinstance(other, type(self)) else False
    
    def __eq__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) == int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() == other.__int__() if isinstance(other, type(self)) else False
    
    def __le__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) <= int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() <= other.__int__() if isinstance(other, type(self)) else False
    
    def __ge__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) >= int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() >= other.__int__() if isinstance(other, type(self)) else False
    
    def __ne__(self, other):
        """
        \\@since 0.3.28
        
        To return true, following conditions must be satisfied:
        - `int(self) != int(other)`
        - `other` must be instance of `RGB` class
        """
        return self.__int__() != other.__int__() if isinstance(other, type(self)) else False
    
    def __pos__(self):
        """
        \\@since 0.3.28
        
        Returns a RGB tuple
        """
        return self.__rgb
    
    def __neg__(self):
        """
        \\@since 0.3.28
        
        Returns a RGB tuple
        """
        return self.__rgb
    
    def __invert__(self):
        """
        \\@since 0.3.28
        
        Returns a RGB tuple
        """
        return self.__rgb
    
    
class CMYK(_util.Final):
    """
    \\@since 0.3.28
    
    ```
    class CMYK(Final): ...
    from tense import CMYK
    ```
    
    Auxiliary class for `Color` class. Represents cyan-magenta-yellow color representation. \\
    Once instantiated, returns `RGB` class instance, only with inverted color values, that is: \\
    255 is 0, 254 is 1, 253 is 2 and so on, up to 0 being 255.
    """
    
    def __new__(self, cyan = 0, magenta = 0, yellow = 0, /):
        
        _parameters = {
            "cyan": cyan,
            "magenta": magenta,
            "yellow": yellow
        }
        
        for key in _parameters:
            
            if not isinstance(_parameters[key], int) or (isinstance(_parameters[key], int) and _parameters[key] not in abroad(0x100)):
                error = TypeError("expected a non-negative integer in parameter '" + key + "' in range 0-255")
                raise error
            
        return RGB(
            0xff - cyan,
            0xff - magenta,
            0xff - yellow
        )
        
class _FileRead(_tc.IntegerFlag):
    "\\@since 0.3.26rc2"
    CHARS = 0
    LINES = _tc.auto()

# _FileReadType = _lit[_FileRead.CHARS, _FileRead.LINES] # unnecessary since 0.3.27b2
_T_stringOrIterable = _var("_T_stringOrIterable", bound = _uni[str, _tc.Iterable[str]])

@_tc.deprecated("Deprecated since 0.3.32, will be removed on 0.3.36")
class File:
    """
    \\@since 0.3.25 \\
    \\@author Aveyzan
    ```ts
    // created 18.07.2024
    in module tense
    ```
    Providing file IO operations
    """
    CHARS = _FileRead.CHARS
    LINES = _FileRead.LINES
    
    def __init__(self, fn: _FileType, mode: _FileMode = "r", /, buffering: int = -1, encoding: _opt[str] = None, errors: _opt[str] = None, newline: _opt[str] = None, closefd: bool = True, opener: _opt[_FileOpener] = None):
        
        self.__i = open(fn, mode, buffering, encoding, errors, newline, closefd, opener)
        
    @_tc.overload
    def read(self, mode: _lit[_FileRead.CHARS, "CHARS", "chars", "_constants", "C"] = CHARS, size: int = -1, /) -> str: ...
    
    @_tc.overload
    def read(self, mode: _lit[_FileRead.LINES, "LINES", "lines", "l", "L"], hint: int = -1, /) -> list[str]: ...
        
    def read(self, mode = CHARS, /, size = -1):
        "\\@since 0.3.26rc2"
        
        if not Tense.isInteger(size):
            error = TypeError("expected 'size' parameter to be an integer")
            raise error
        
        if self.__i.readable():
            
            if mode in (self.CHARS, "CHARS", "chars", "c", "C"):
                return self.__i.read(size)
            
            elif mode in (self.LINES, "LINES", "lines", "l", "L"):
                return self.__i.readlines(size)
            
            else:
                error = TypeError("expected one of constants: 'CHARS', 'LINES'")
                raise error
            
        else:
            error = IOError("file is not open for reading")
            raise error
        
    def write(self, content: _T_stringOrIterable):
        "\\@since 0.3.26rc2"
        
        if not Tense.isString(content) and not isinstance(content, _tc.Iterable):
            error = TypeError("Expected a string iterable or a string")
            raise error
    
        if self.__i.writable():
            
            if Tense.isString(content):
                self.__i.write(content)
                
            else:
                self.__i.writelines(content)
                
        else:
            error = IOError("file is not open for writing")
            raise error
        
    def pickle(self, o: object, protocol: _opt[int] = None, *, fixImports = True, bufferCallback: _opt[_cal[[_pickle.PickleBuffer], None]] = None):
        "\\@since 0.3.26rc2"
        if isinstance(self.__i, (
            # only on binary mode files
            _io.BufferedRandom,
            _io.BufferedReader,
            _io.BufferedWriter
        )):
            _pickle.dump(o, self.__i, protocol, fix_imports = fixImports, buffer_callback = bufferCallback)
            
        else:
            error = IOError("file is not open in binary mode")
            raise error
        
    def unpickle(self, *, fixImports = True, encoding = "ASCII", errors = "strict", buffers: _opt[_tc.Iterable[_Any]] = ()):
        "\\@since 0.3.26rc2"
        
        a = []
        
        if isinstance(self.__i, (
            # only on binary mode files
            _io.BufferedRandom,
            _io.BufferedReader,
            _io.BufferedWriter
        )):
            
            while True:
            
                try:
                    a.append(_pickle.load(self.__i, fix_imports = fixImports, encoding = encoding, errors = errors, buffers = buffers))
                    
                except:
                    break
                
            return a
            
        else:
            error = (IOError, "file is not open in binary mode")
            raise error
        

class FinalVar:
    """
    \\@since 0.3.26rc1 \\
    \\@lifetime ≥ 0.3.26rc1
    ```py
    class FinalVar: ...
    from tense[.types_collection] import FinalVar
    ```
    A class-like function (since 0.3.26rc2 a class) referring \\
    to class `tense.types_collection.FinalVar`. This creates a final variable

    Use `~instance`, `+instance` or `-instance` to return \\
    value passed to the constructor.

    To allow it work in classes, you need to initialize them to obtain the value. \\
    In case of variables in global scope it works normally.
    """

    def __new__(cls, v: _T, /):
        """
        \\@since 0.3.26rc1 \\
        \\@lifetime ≥ 0.3.26rc1
        ```
        in module tense
        ```
        A class-like function (since 0.3.26rc2 a class) referring \\
        to class `tense.types_collection.FinalVar`. This creates a final variable

        Use `~instance`, `+instance` or `-instance` to return \\
        value passed to the constructor.

        To allow it work in classes, you need to initialize them to obtain the value. \\
        In case of variables in global scope it works normally.
        """
        return _tc.FinalVar(v)
    
class ClassLike:
    """
    \\@since 0.3.27a3 \\
    \\@lifetime ≥ 0.3.27a3
    ```
    in module tense
    ```
    A class being formal reference to class `tense.types_collection.ClassLike`
    """

    def __new__(cls, f: _tc.Callable[_P, _T]):
        """
        \\@since 0.3.27a3 \\
        \\@lifetime ≥ 0.3.27a3
        ```
        in module tense
        ```
        A class being formal reference to class `tense.types_collection.ClassLike`
        """
        return _tc.ClassLike(f)

# class _ChangeVarState(tc.IntegerFlag): # to 0.3.28
class _ChangeVarState(_tc.Enum):
    "\\@since 0.3.26rc1. Internal class for `ChangeVar.setState()` method"
    I = 1
    D = 2

# _ChangeVarStateSelection = _lit[_ChangeVarState.D, _ChangeVarState.I] # unnecessary since 0.3.27b2

class ChangeVar(_tc.UnaryOperable, _tc.Comparable, _tc.AdditionReassignable, _tc.SubtractionReassignable):
    """
    \\@since 0.3.26rc1 \\
    \\@lifetime ≥ 0.3.26rc1
    ```py
    class ChangeVar: ...
    from tense import ChangeVar
    ```
    Auxiliary class for creating sentinel inside `while` loop.

    Use `~instance` to receive integer value. \\
    Use `+instance` to increment by 1. \\
    Use `-instance` to decrement by 1. \\
    Use `instance += any_int` to increment by `any_int`. \\
    Use `instance -= any_int` to decrement by `any_int`.
    """
    D = _ChangeVarState.D
    I = _ChangeVarState.I
    __v = 0
    __m = 1
    __default = 0

    def __init__(self, initialValue = 0):
        
        if not Tense.isInteger(initialValue):
            error = TypeError("expected an integer value")
            raise error
        
        self.__v = initialValue
        self.__default = initialValue

    def __pos__(self):
        self.__v += self.__m

    def __neg__(self):
        self.__v -= self.__m

    def __invert__(self):
        return self.__v
    
    def __eq__(self, other: int):
        return self.__v == other
    
    def __contains__(self, value: int):
        return self.__v == value
    
    def __ne__(self, other: int):
        return self.__v != other
    
    def __ge__(self, other: int):
        return self.__v >= other
    
    def __gt__(self, other: int):
        return self.__v > other
    
    def __le__(self, other: int):
        return self.__v <= other
    
    def __lt__(self, other: int):
        return self.__v < other
    
    def __iadd__(self, other: int):
        
        if not Tense.isInteger(other):
            error = TypeError("expected an integer as a right operand") # error replaced 0.3.34; earlier was NotInitializedError
            raise error
        
        _tmp = self.__v
        _tmp += other
        self.__v = _tmp
        return _tmp
    
    def __isub__(self, other: int):
        
        if not Tense.isInteger(other):
            error = TypeError("expected an integer as a right operand") # error replaced 0.3.34; earlier was NotInitializedError
            raise error
        
        _tmp = self.__v
        _tmp -= other
        self.__v = _tmp
        return _tmp
    
    def reset(self):
        """
        \\@since 0.3.26rc1

        Reset the counter to value passed to the constructor, or - \\
        if `setDefault()` was invoked before - to value passed \\
        to that method.
        """
        self.__v = self.__default

    def setDefault(self, value: int):
        """
        \\@since 0.3.26rc1

        Set a new default value. This overwrites current default value. \\
        Whether `reset()` method is used after, internal variable \\
        will have the default value, which was passed to this method. \\
        Otherwise it will refer to value passed to constructor
        """
        if not Tense.isInteger(value):
            error = TypeError("expected an integer value")
            raise error
        self.__default = abs(value)

    def setState(self, s: _ChangeVarState = I, m: int = 1):
        """
        \\@since 0.3.26rc1

        Alternative for `+` and `-` unary operators.

        If `D` for `s` parameter is passed, sentinel will be decremented \\
        by 1, otherwise incremented by 1 (option `I`). Additionally, you \\
        can set a different step via `m` parameter.
        """
        _m = m
        
        if not Tense.isInteger(_m):
            err, _s = TypeError("expected integer value for 'm' parameter")
            raise err(_s)
        
        elif abs(_m) == 0:
            _m = 1
            
        if s == self.D:
            self.__v -= abs(_m)
            
        elif s == self.I:
            self.__v += abs(_m)
            
        else:
            err, _s = TypeError("expected 'ChangeVar.I' or 'ChangeVar.D' for 's' parameter")
            raise err(_s)
        
    def setModifier(self, m: int):
        """
        \\@since 0.3.26rc1

        Changes behavior for `+` and `-` unary operators. \\
        If passed integer value was negative, code will \\
        retrieve absolute value of it. If 0 passed, used will be 1
        """
        _params = [k for k in self.setModifier.__annotations__]
        
        if not Tense.isInteger(m):
            error = TypeError("expected integer value in parameter '{}'".format(_params[0]))
            raise error
        
        elif abs(m) == 0:
            self.__m == 1
            
        self.__m = abs(m)

_Color = _tc.Union[_tc.ColorType, RGB]

# class _ColorStyling(_tc.IntegerFlag): ### to 0.3.27
class _ColorStyling(_tc.Enum):
    "\\@since 0.3.26rc1. Internal class for `%` operator in class `tense.Color`."
    NORMAL = 0
    BOLD = 1
    FAINT = 2
    ITALIC = 3
    UNDERLINE = 4
    SLOW_BLINK = 5
    RAPID_BLINK = 6
    REVERSE = 7
    HIDE = 8
    STRIKE = 9
    # PRIMARY_FONT = 10
    ## 11-19 alternative font
    # GOTHIC = 20
    DOUBLE_UNDERLINE = 21
    # NORMAL_INTENSITY = 22
    # NO_ITALIC = 23
    # NO_UNDERLINE = 24
    # NO_BLINK = 25
    # PROPORTIONAL = 26 # corrected mistake! 0.3.26rc2
    # NO_REVERSE = 27
    # UNHIDE = 28
    # NO_STRIKE = 29
    ## 30-37 foreground color, 3-bit
    # 38 foreground color, 3 4 8 24-bit
    # FOREGROUND_DEFAULT = 39
    ## 40-47 background color, 3-bit
    ## 48 background color, 3 4 8 24-bit
    # BACKGROUND_DEFAULT = 49
    # NO_PROPORTIONAL = 50
    FRAME = 51
    ENCIRCLE = 52
    OVERLINE = 53
    # NO_FRAME = 54 # including "no encircle"
    # NO_OVERLINE = 55
    ## 56 and 57 undefined
    ## 58 underline color, 3 4 8 24-bit
    # UNDERLINE_DEFAULT = 59
    # IDEOGRAM_UNDERLINE = 60
    # IDEOGRAM_DOUBLE_UNDERLINE = 61
    # IDEOGRAM_OVERLINE = 62
    # IDEOGRAM_DOUBLE_OVERLINE = 63
    # IDEOGRAM_STRESS = 64
    # NO_IDEOGRAM = 65
    ## 66-72 undefined
    SUPERSCRIPT = 73
    SUBSCRIPT = 74
    # NO_SUPERSCRIPT = 75 # also counts as no subscript
    ## 76 undefined but recommended value: no subscript
    ## 77-89 undefined
    ## 90-97 bright foreground color, 4-bit
    ## 100-107 bright background color, 4-bit

# class _ColorAdvancedStyling(_tc.IntegerFlag): ### to 0.3.27
class _ColorAdvancedStyling(_tc.Enum):
    "\\@since 0.3.26rc2. Internal class for `%` operator in class `tense.Color`."
    # 2x
    BOLD_ITALIC = 1000
    BOLD_UNDERLINE = 1001
    BOLD_STRIKE = 1002
    BOLD_OVERLINE = 1003
    ITALIC_UNDERLINE = 1004
    ITALIC_STRIKE = 1005
    ITALIC_OVERLINE = 1006
    UNDERLINE_STRIKE = 1007
    UOLINE = 1008
    STRIKE_OVERLINE = 1009
    # 3x
    BOLD_ITALIC_UNDERLINE = 1100
    BOLD_ITALIC_STRIKE = 1101
    BOLD_ITALIC_OVERLINE = 1102
    BOLD_UNDERLINE_STRIKE = 1103
    BOLD_UOLINE = 1104
    ITALIC_UNDERLINE_STRIKE = 1105
    ITALIC_UOLINE = 1106
    ITALIC_STRIKE_OVERLINE = 1107
    STRIKE_UOLINE = 1108

_ColorStylingType = _uni[_ColorStyling, _ColorAdvancedStyling, _lit["normal", "bold", "faint", "italic", "underline", "slow_blink", "rapid_blink", "reverse", "hide", "strike", "double_underline", "frame", "encircle", "overline", "superscript", "subscript"]]

class Color(_tc.ModuloOperable[_ColorStylingType, str], _tc.UnaryOperable):
    """
    \\@since 0.3.26rc1 \\
    \\@lifetime ≥ 0.3.26rc1
    ```py
    class Color: ...
    from tense import Color
    ```
    Deputy of experimental class `tense.extensions.ANSIColor` (≥ 0.3.24; < 0.3.26rc1).
    
    This class uses ANSI escape code for color purposes.

    Unary `+` and `-`, and `~` operators allow to get colored string. \\
    Since 0.3.34 using `str()` will also return colored string, other 3 ways \\
    to do so will remain due to compatibility with older versions of Tense.
    
    Modulo operator (`%`) allows to change the font style. The right operand must be \\
    an appropriate constant or lowercased string literal (for single font styles).
    Examples:
    ```ts
    Color("Tense") % Color.BOLD
    Color("Countryside!", 8, 0o105) % Color.ITALIC // italic, blue text
    Color("Creativity!", 24, 0xc0ffee) % Color.BOLD // bold, c0ffee hex code text
    Color("Illusive!", 24, 0, 0xc0ffee) % Color.BOLD // bold, c0ffee hex code background, black text
    ```

    Since 0.3.26rc2 you can use constants, which grant more than one font style simultaneously, like:
    ```ts
    Color("Lines!", 8, 93) % Color.UOLINE // lines above and below text
    ```

    **Warning**: 24-bit colors load longer than colors from lower bit shelves. In this case it is \\
    recommended to stick to 8-bit colors, but if there isn't a satisfying color, 24-bit color support \\
    will be kept. It is also somewhat a reason of `RGB` and `CMYK` colors existence.
    """
    __fg = None
    __bg = None
    if False: # 0.3.27
        __un = None
    __text = ""
    __bits = 8 # 24 to 0.3.34 

    NORMAL = _ColorStyling.NORMAL
    "\\@since 0.3.26rc1. Mere text"
    
    BOLD = _ColorStyling.BOLD
    "\\@since 0.3.26rc1. Text becomes bold"
    
    FAINT = _ColorStyling.FAINT
    "\\@since 0.3.26rc1. Also works as 'decreased intensity' or 'dim'"
    
    ITALIC = _ColorStyling.ITALIC
    "\\@since 0.3.26rc1. Text becomes oblique. Not widely supported"
    
    UNDERLINE = _ColorStyling.UNDERLINE
    "\\@since 0.3.26rc1. Text becomes underlined. Marked *experimental* as experimenting with underline colors, but normally it is OK to use"
    
    SLOW_BLINK = _ColorStyling.SLOW_BLINK
    "\\@since 0.3.26rc1. Text will blink for less than 150 times per minute"
    
    RAPID_BLINK = _ColorStyling.RAPID_BLINK
    "\\@since 0.3.26rc1. Text will blink for more than 150 times per minute. Not widely supported"
    
    REVERSE = _ColorStyling.REVERSE
    "\\@since 0.3.26rc2. Swap text and background colors"
    
    HIDE = _ColorStyling.HIDE
    "\\@since 0.3.26rc1. Text becomes transparent"
    
    STRIKE = _ColorStyling.STRIKE
    "\\@since 0.3.26rc1. Text becomes crossed out"
    
    DOUBLE_UNDERLINE = _ColorStyling.DOUBLE_UNDERLINE
    "\\@since 0.3.26rc2. Text becomes doubly underlined"
    
    # PROPORTIONAL = _ColorStyling.PROPORTIONAL
    "\\@lifetime >= 0.3.26rc1; < 0.3.26rc2. Proportional spacing. *Experimental*"
    
    FRAME = _ColorStyling.FRAME
    "\\@since 0.3.26rc1. Implemented in mintty as 'emoji variation selector'"
    
    ENCIRCLE = _ColorStyling.ENCIRCLE
    "\\@since 0.3.26rc1. Implemented in mintty as 'emoji variation selector'"
    
    OVERLINE = _ColorStyling.OVERLINE
    "\\@since 0.3.26rc1. Text becomes overlined"
    
    SUPERSCRIPT = _ColorStyling.SUPERSCRIPT
    "\\@since 0.3.26rc2. Text becomes superscripted (implemented in mintty only)"
    
    SUBSCRIPT = _ColorStyling.SUBSCRIPT
    "\\@since 0.3.26rc2. Text becomes subscripted (implemented in mintty only)"
    
    # 2x
    BOLD_ITALIC = _ColorAdvancedStyling.BOLD_ITALIC
    "\\@since 0.3.26rc2. Text becomes bold and oblique"
    
    BOLD_UNDERLINE = _ColorAdvancedStyling.BOLD_UNDERLINE
    "\\@since 0.3.26rc2. Text becomes bold and underlined"
    
    BOLD_STRIKE = _ColorAdvancedStyling.BOLD_STRIKE
    "\\@since 0.3.26rc2. Text becomes bold and crossed out"
    
    BOLD_OVERLINE = _ColorAdvancedStyling.BOLD_OVERLINE
    "\\@since 0.3.26rc2. Text becomes bold and overlined"
    
    ITALIC_UNDERLINE = _ColorAdvancedStyling.ITALIC_UNDERLINE
    "\\@since 0.3.26rc2. Text becomes oblique and underlined"
    
    ITALIC_STRIKE = _ColorAdvancedStyling.ITALIC_STRIKE
    "\\@since 0.3.26rc2. Text becomes oblique and crossed out"
    
    ITALIC_OVERLINE = _ColorAdvancedStyling.ITALIC_OVERLINE
    "\\@since 0.3.26rc2. Text becomes oblique and overlined"
    
    UNDERLINE_STRIKE = _ColorAdvancedStyling.UNDERLINE_STRIKE
    "\\@since 0.3.26rc2. Text becomes underlined and crossed out"
    
    UOLINE = _ColorAdvancedStyling.UOLINE
    "\\@since 0.3.26rc2. Alias to underline-overline. Text gets lines above and below"
    
    STRIKE_OVERLINE = _ColorAdvancedStyling.STRIKE_OVERLINE
    "\\@since 0.3.26rc2. Text becomes crossed out and overlined"
    
    # 3x
    BOLD_ITALIC_UNDERLINE = _ColorAdvancedStyling.BOLD_ITALIC_UNDERLINE
    "\\@since 0.3.26rc2. Text becomes bold, oblique and underlined"
    
    BOLD_ITALIC_STRIKE = _ColorAdvancedStyling.BOLD_ITALIC_STRIKE
    "\\@since 0.3.26rc2"
    
    BOLD_ITALIC_OVERLINE = _ColorAdvancedStyling.BOLD_ITALIC_OVERLINE
    "\\@since 0.3.26rc2"
    
    BOLD_UNDERLINE_STRIKE = _ColorAdvancedStyling.BOLD_UNDERLINE_STRIKE
    "\\@since 0.3.26rc2"
    
    BOLD_UOLINE = _ColorAdvancedStyling.BOLD_UOLINE
    "\\@since 0.3.26rc2"
    
    ITALIC_UNDERLINE_STRIKE = _ColorAdvancedStyling.ITALIC_UNDERLINE_STRIKE
    "\\@since 0.3.26rc2"
    
    ITALIC_UOLINE = _ColorAdvancedStyling.ITALIC_UOLINE
    "\\@since 0.3.26rc2"
    
    ITALIC_STRIKE_OVERLINE = _ColorAdvancedStyling.ITALIC_STRIKE_OVERLINE
    "\\@since 0.3.26rc2"
    
    STRIKE_UOLINE = _ColorAdvancedStyling.STRIKE_UOLINE
    "\\@since 0.3.26rc2"
        
    def __prepare_return(self):
        
        _s = "\033["
        # for e in (self.__fg, self.__bg, self.__un): ### removed 0.3.27
        _msg = {
            3: "for 3-bit colors, expected integer or string value in range 0-7. One of foreground or background values doesn't match this requirement",
            4: "for 4-bit colors, expected integer or string value in range 0-15. One of foreground or background values doesn't match this requirement",
            8: "for 8-bit colors, expected integer or string value in range 0-255. One of foreground or background values doesn't match this requirement",
            24: "for 2-bit colors, expected integer, string, or RGB/CMYK tuple value in range 0-16777215. One of foreground or background values doesn't match this requirement"
        }
        
        for e in (self.__fg, self.__bg):
            
            if e is not None:
                if self.__bits == 3 and e not in abroad(0x8):
                    error = ValueError(_msg[3])
                    raise error
                
                elif self.__bits == 4 and e not in abroad(0x10):
                    error = ValueError(_msg[4])
                    raise error
                
                elif self.__bits == 8 and e not in abroad(0x100):
                    error = ValueError(_msg[8])
                    raise error
            
                elif self.__bits == 24 and e not in abroad(0x1000000):
                    error = ValueError(_msg[24])
                    raise error
        
        if self.__bits == 3:
            # 2 ** 3 = 8 (0x8 in hex)
            _s += str(30 + self.__fg) + ";" if self.__fg is not None else ""
            _s += str(40 + self.__bg) + ";" if self.__bg is not None else ""
            # _s += "58;5;" + str(self.__un) + ";" if self.__un is not None else "" ### removed 0.3.27
        
        elif self.__bits == 4:
            # 2 ** 4 = 16 (0x10 in hex); WARNING: bright colors notation isn't official
            _s += str(30 + self.__fg) + ";" if self.__fg is not None and self.__fg in abroad(0x8) else ""
            _s += str(40 + self.__bg) + ";" if self.__bg is not None and self.__bg in abroad(0x8) else ""
            _s += str(90 + self.__fg) + ";" if self.__fg is not None and self.__fg in abroad(0x8, 0x10) else ""
            _s += str(100 + self.__bg) + ";" if self.__bg is not None and self.__bg in abroad(0x8, 0x10) else ""
            # _s += "58;5;" + str(self.__un) + ";" if self.__un is not None else "" ### removed 0.3.27
        
        elif self.__bits == 8:
            # 2 ** 8 = 256 (0x100 in hex)
            _s += "38;5;" + str(self.__fg) + ";" if self.__fg is not None else ""
            _s += "48;5;" + str(self.__bg) + ";" if self.__bg is not None else ""
            # _s += "58;5;" + str(self.__un) + ";" if self.__un is not None else "" ### removed 0.3.27
        
        elif self.__bits == 24:
            # 2 ** 24 = 16777216 (0x1000000 in hex)
            # code reconstructed on 0.3.26rc2
            # acknowledgements: equivalent to rgb
            _f = hex(self.__fg) if self.__fg is not None else ""
            _b = hex(self.__bg) if self.__bg is not None else ""
            # _u = hex(self.__un) if self.__un is not None else "" ### removed 0.3.27
            _f = _re.sub(r"^(0x|#)", "", _f) if reckon(_f) > 0 else ""
            _b = _re.sub(r"^(0x|#)", "", _b) if reckon(_b) > 0 else ""
            # _u = _re.sub(r"^(0x|#)", "", _u) if reckon(_u) > 0 else "" ### removed 0.3.27
            # _hf, _hb, _hu = [None for _ in abroad(3)] ### removed 0.3.27
            _hf, _hb = [None, None]
            # for s in (_f, _b, _u): ### removed 0.3.27
            for s in (_f, _b):
                
                if reckon(s) == 6:
                    if s == _f:
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2)) ### removed 0.3.27
                    
                elif reckon(s) == 5:
                    s = "0" + s
                    if s == "0" + _f:
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    
                elif reckon(s) == 4:
                    s = "00" + s
                    if s == "00" + _f:
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2)) ### removed 0.3.27
                    
                elif reckon(s) == 3:
                    _tmp = "".join(s[i] * 2 for i in abroad(s)) # aliased according to css hex fff notation
                    if s == _f:
                        s = _tmp
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        s = _tmp
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    s = _tmp
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2)) ### removed 0.3.27
                    
                elif reckon(s) == 2:
                    s = "0000" + s
                    if s == "0000" + _f:
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2)) ### removed 0.3.27
                    
                elif reckon(s) == 1:
                    s = "00000" + s
                    if s == "00000" + _f:
                        _hf = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    else:
                        _hb = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2))
                    # else:
                    #    _hu = tuple(int(s[i : i + 2], 16) for i in abroad(0, 5, 2)) ### removed 0.3.27
            
            _s += "38;2;" + str(_hf[0]) + ";" + str(_hf[1]) + ";" + str(_hf[2]) + ";" if _hf is not None else ""
            _s += "48;2;" + str(_hb[0]) + ";" + str(_hb[1]) + ";" + str(_hb[2]) + ";" if _hb is not None else ""
            # _s += "58;2;" + str(_hu[0]) + ";" + str(_hu[1]) + ";" + str(_hu[2]) + ";" if _hu is not None else "" ### removed 0.3.27
        else:
            error = ValueError("internal 'bits' variable value is not one from following: 3, 4, 8, 24")
            raise error
        
        if _s != "\033[":
            _s = _re.sub(r";$", "m", _s)
            _s += self.__text + "\033[0m"
        else:
            _s = self.__text
        return _s
    
    if True: # since 0.3.27
        
        def __init__(self, text: str, /, bits: _Bits = 8, foregroundColor: _Color = None, backgroundColor: _Color = None): # slash since 0.3.26rc2
            """
            \\@since 0.3.26rc1. Parameters:
            - `text` - string to be colored. Required parameter
            - `bits` - number of bits, possible values: 3, 4, 8, 24. Defaults to 24 (since 0.3.26rc2 - 8)
            - `foregroundColor` - color of the foreground (text). String/integer/`None`. Defaults to `None`
            - `backgroundColor` - color of the background. String/integer/`None`. Defaults to `None`
            
            See https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit for color palette outside 24-bit colors
            """
            _os.system("color")
            
            _params = _get_all_params(self.__init__)
            
            if not Tense.isString(text):
                error = TypeError("expected string value for '{}' parameter".format(_params[0]))
                raise error
            
            if not Tense.isInteger(bits) or (Tense.isInteger(bits) and bits not in (3, 4, 8, 24)):
                error = TypeError("expected integer value: 3, 4, 8 or 24, for '{}' parameter".format(_params[1]))
                raise error
            
            for e in (foregroundColor, backgroundColor):
                
                if not Tense.isInteger(e) and not Tense.isString(e) and not isinstance(e, RGB) and not Tense.isNone(e):
                    error = TypeError("expected integer, string or 'None' value for both foreground and background color parameters")
                    raise error
                
                elif Tense.isString(e) and (
                    # changing this order may cause easier error
                    not _is_hexadecimal(e) and
                    not _is_decimal(e) and
                    not _is_octal(e) and
                    not _is_binary(e)
                ):
                    error = TypeError(f"malformed string in either foreground or background color parameters, expected clean binary, decimal, hexademical or octal string")
                    raise error
                
                elif bits == 24 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x1000000) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x1000000) or
                    isinstance(e, RGB) and int(e) not in abroad(0x1000000)
                ):
                    error = ValueError(f"for 24-bit colors, expected \"RGB\" class instance of integer value, integer or string value in range 0-16777215")
                    raise error
                
                elif bits == 8 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x100) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x100) or isinstance(e, RGB)
                ):
                    error = ValueError(f"for 8-bit colors, expected integer or string value in range 0-255. Cannot be used with \"RGB\" class instance")
                    raise error
                
                elif bits == 4 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x10) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x10) or isinstance(e, RGB)
                ):
                    error = ValueError(f"for 4-bit colors, expected integer or string value in range 0-15. Cannot be used with \"RGB\" class instance")
                    raise error
                
                elif bits == 3 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x8) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x8) or isinstance(e, RGB)
                ):
                    error = ValueError(f"for 3-bit colors, expected integer or string value in range 0-7. Cannot be used with \"RGB\" class instance")
                    raise error
            
            self.__text = text
            self.__bits = bits
            self.__fg = foregroundColor if Tense.isInteger(foregroundColor) else _int_conversion(foregroundColor) if Tense.isString(foregroundColor) else int(foregroundColor) if isinstance(foregroundColor, RGB) else None
            self.__bg = backgroundColor if Tense.isInteger(backgroundColor) else _int_conversion(backgroundColor) if Tense.isString(backgroundColor) else int(backgroundColor) if isinstance(backgroundColor, RGB) else None
            
    else:
        def __init__(self, text: str, /, bits: _Bits = 8, foregroundColor: _Color = None, backgroundColor: _Color = None, underlineColor: _Color = None):
            
            _os.system("color")
            
            if not Tense.isString(text):
                error = TypeError("Expected string value for 'text' parameter")
                raise error
            
            if not Tense.isInteger(bits) or (Tense.isInteger(bits) and bits not in (3, 4, 8, 24)):
                error = TypeError("Expected integer value: 3, 4, 8 or 24, for 'bits' parameter")
                raise error
            
            for e in (foregroundColor, backgroundColor, underlineColor):
                
                if not Tense.isInteger(e) and not Tense.isString(e) and e is not None:
                    error = TypeError(f"Expected integer, string or 'None' value for '{e.__name__}' parameter")
                    raise error
                
                elif Tense.isString(e) and (
                    not _is_hexadecimal(e) and
                    not _is_decimal(e) and
                    not _is_octal(e) and
                    not _is_binary(e)
                ):
                    error = TypeError(f"Malformed string in parameter 'e', expected clean binary, decimal, hexademical or octal string")
                    raise error
                
                elif bits == 24 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x1000000) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x1000000)
                ):
                    error = ValueError(f"For 24-bit colors, expected integer or string value in range 0-16777215")
                    raise error
                
                elif bits == 8 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x100) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x100)
                ):
                    error = ValueError(f"For 8-bit colors, expected integer or string value in range 0-255")
                    raise error
                
                elif bits == 4 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x10) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x10)
                ):
                    error = ValueError(f"For 4-bit colors, expected integer or string value in range 0-15")
                    raise error
                
                elif bits == 3 and e is not None and (
                    Tense.isInteger(e) and e not in abroad(0x8) or
                    Tense.isString(e) and _int_conversion(e) not in abroad(0x8)
                ):
                    error = ValueError(f"For 3-bit colors, expected integer or string value in range 0-7")
                    raise error
                
            self.__text = text
            self.__bits = bits
            self.__fg = foregroundColor if Tense.isInteger(foregroundColor) else _int_conversion(foregroundColor) if Tense.isString(foregroundColor) else None
            self.__bg = backgroundColor if Tense.isInteger(backgroundColor) else _int_conversion(backgroundColor) if Tense.isString(backgroundColor) else None
            self.__un = underlineColor if Tense.isInteger(underlineColor) else _int_conversion(underlineColor) if Tense.isString(underlineColor) else None
    
    def clear(self):
        """
        \\@since 0.3.26rc1
        
        Clear every color for foreground, background and underline. Should \\
        be used before `setBits()` method invocation to avoid conflicts. \\
        By default bits value is reset to 24. Since 0.3.27b1 - 8.
        """
        self.__fg = None
        self.__bg = None
        if False: # 0.3.27
            self.__un = None
        self.__bits = 8
        return self
    
    def setBits(self, bits: _Bits = 8, /):
        """
        \\@since 0.3.26rc1

        Possible values: 3, 4, 8, 24. Default is 24. \\
        Since 0.3.26rc2 default value is 8.
        """
        
        if not Tense.isInteger(bits) or (Tense.isInteger(bits) and bits not in (3, 4, 8, 24)):
            error = TypeError("expected integer value: 3, 4, 8 or 24, for 'bits' parameter")
            raise error
        
        # for e in (self.__fg, self.__bg, self.__un): ### removed 0.3.27
        for e in (self.__fg, self.__bg):
            
            if e is not None:
                
                if bits == 24 and e not in abroad(0x1000000):
                    error = ValueError("internal conflict caught while setting 'bits' value to 24. One of foreground or background values is beyond range 0-16777215. To prevent this conflict, use method 'Color.clear()'.")
                    raise error
                
                elif bits == 8 and e not in abroad(0x100):
                    error = ValueError("internal conflict caught while setting 'bits' value to 8. One of foreground or background values is beyond range 0-255. To prevent this conflict, use method 'Color.clear()'.")
                    raise error
                
                elif bits == 4 and e not in abroad(0x10):
                    error = ValueError("internal conflict caught while setting 'bits' value to 4. One of foreground or background values is beyond range 0-15. To prevent this conflict, use method 'Color.clear()'.")
                    raise error
                
                elif bits == 3 and e not in abroad(0x8):
                    error = ValueError("internal conflict caught while setting 'bits' value to 3. One of foreground or background values is beyond range 0-7. To prevent this conflict, use method 'Color.clear()'.")
                    raise error
                
        self.__bits = bits
    
    def setForegroundColor(self, color: _Color = None, /):
        """
        \\@since 0.3.26rc1
        
        Set foreground color manually.
        """
        _c = color if Tense.isInteger(color) or color is None else _int_conversion(color) if Tense.isString(color) else int(color) if isinstance(color, RGB) else None
        
        if _c is not None:
            
            if self.__bits == 3 and _c not in abroad(0x8):
                error = ValueError(f"for 3-bit colors, expected integer or string value in range 0-7")
                raise error
            
            elif self.__bits == 4 and _c not in abroad(0x10):
                error = ValueError(f"for 4-bit colors, expected integer or string value in range 0-15")
                raise error
            
            elif self.__bits == 8 and _c not in abroad(0x100):
                error = ValueError(f"for 8-bit colors, expected integer or string value in range 0-255")
                raise error
            
            elif self.__bits == 24 and _c not in abroad(0x1000000):
                error = ValueError(f"for 24-bit colors, expected integer, string or RGB/CMYK tuple value in range 0-16777215")
                raise error
            
            else:
                error = ValueError(f"internal 'bits' variable value is not one from following: 3, 4, 8, 24")
                raise error
            
        self.__fg = _c
        return self
    
    def setBackgroundColor(self, color: _Color = None, /):
        """
        \\@since 0.3.26rc1
        
        Set background color manually.
        """
        _c = color if Tense.isInteger(color) or color is None else _int_conversion(color) if Tense.isString(color) else int(color) if isinstance(color, RGB) else None
        
        if _c is not None:
            
            if self.__bits == 3 and _c not in abroad(0x8):
                error = ValueError(f"for 3-bit colors, expected integer or string value in range 0-7")
                raise error
            
            elif self.__bits == 4 and _c not in abroad(0x10):
                error = ValueError(f"for 4-bit colors, expected integer or string value in range 0-15")
                raise error
            
            elif self.__bits == 8 and _c not in abroad(0x100):
                error = ValueError(f"for 8-bit colors, expected integer or string value in range 0-255")
                raise error
            
            elif self.__bits == 24 and _c not in abroad(0x1000000):
                error = ValueError(f"for 24-bit colors, expected integer, string or RGB/CMYK tuple value in range 0-16777215")
                raise error
            
            else:
                error = ValueError(f"internal 'bits' variable value is not one from following: 3, 4, 8, 24")
                raise error
            
        self.__bg = _c
        return self
    
    if False:
        def setUnderlineColor(self, color: _Color = None, /):
            """
            \\@since 0.3.26rc1
            
            Set underline color manually. *Experimental* \\
            Since 0.3.26rc2 only accepted value is `None`.
            """
            _c = color if Tense.isInteger(color) or color is None else _int_conversion(color)
            if _c is not None:
                if self.__bits == 3 and _c not in abroad(0x8):
                    error = ValueError(f"For 3-bit colors, expected integer or string value in range 0-7")
                    raise error
                
                elif self.__bits == 4 and _c not in abroad(0x10):
                    error = ValueError(f"For 4-bit colors, expected integer or string value in range 0-15")
                    raise error
                
                elif self.__bits == 8 and _c not in abroad(0x100):
                    error = ValueError(f"For 8-bit colors, expected integer or string value in range 0-255")
                    raise error
                
                elif self.__bits == 24 and _c not in abroad(0x1000000):
                    error = ValueError(f"For 24-bit colors, expected integer or string value in range 0-16777215")
                    raise error
                
                else:
                    error = ValueError(f"Internal 'bits' variable value is not one from following: 3, 4, 8, 24")
                    raise error
                
            self.__un = _c
            return self
    
    
    def __pos__(self):
        """\\@since 0.3.26rc1. Receive colored string"""
        return self.__prepare_return()
    
    def __neg__(self):
        """\\@since 0.3.26rc1. Receive colored string"""
        return self.__prepare_return()
    
    def __invert__(self):
        """\\@since 0.3.26rc1. Receive colored string"""
        return self.__prepare_return()
    
    def __str__(self):
        """\\@since 0.3.34. Receive colored string"""
        return self.__prepare_return()
    
    def __mod__(self, other: _ColorStylingType):
        """
        \\@since 0.3.26rc1
        
        Further styling. Use constant, which is in `__constants__` attribute.
        """
        # below: since 0.3.26rc1
        if other in (self.NORMAL, "normal"):
            return self.__prepare_return()
        
        elif other in (self.BOLD, "bold"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;", self.__prepare_return())
        
        elif other in (self.FAINT, "faint"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[2;", self.__prepare_return())
        
        elif other in (self.ITALIC, "italic"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;", self.__prepare_return())
        
        elif other in (self.UNDERLINE, "underline"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[4;", self.__prepare_return())
        
        elif other in (self.SLOW_BLINK, "slow_blink"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[5;", self.__prepare_return())
        
        elif other in (self.RAPID_BLINK, "rapid_blink"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[6;", self.__prepare_return())
        
        # below: since 0.3.26rc2
        elif other in (self.REVERSE, "reverse"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[7;", self.__prepare_return())
        
        # below: since 0.3.26rc1
        elif other in (self.HIDE, "hide"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[8;", self.__prepare_return())
        
        elif other in (self.STRIKE, "strike"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[9;", self.__prepare_return())
        
        elif other in (self.DOUBLE_UNDERLINE, "double_underline"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[21;", self.__prepare_return())
        
        # elif other == self.PROPORTIONAL:
        #    return _re.sub(r"^(\033\[|\u001b\[)", "\033[26;", self.__prepare_return())
        
        elif other in (self.FRAME, "frame"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[51;", self.__prepare_return())
        
        elif other in (self.ENCIRCLE, "encircle"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[52;", self.__prepare_return())
        
        elif other in (self.OVERLINE, "overline"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[53;", self.__prepare_return())
        
        # below: since 0.3.26rc2
        elif other in (self.SUPERSCRIPT, "superscript"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[73;", self.__prepare_return())
        
        elif other in (self.SUBSCRIPT, "subscript"):
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[74;", self.__prepare_return())
        # 2x; since 0.3.26rc2
        elif other == self.BOLD_ITALIC:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;3;", self.__prepare_return())
        
        elif other == self.BOLD_UNDERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;4;", self.__prepare_return())
        
        elif other == self.BOLD_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;9;", self.__prepare_return())
        
        elif other == self.BOLD_OVERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;53;", self.__prepare_return())
        
        elif other == self.ITALIC_UNDERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;4;", self.__prepare_return())
        
        elif other == self.ITALIC_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;9;", self.__prepare_return())
        
        elif other == self.ITALIC_OVERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;53;", self.__prepare_return())
        
        elif other == self.UNDERLINE_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[4;9;", self.__prepare_return())
        
        elif other == self.UOLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[4;53;", self.__prepare_return())
        
        elif other == self.STRIKE_OVERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[9;53;", self.__prepare_return())
        
        # 3x; since 0.3.26rc2
        elif other == self.BOLD_ITALIC_UNDERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;3;4;", self.__prepare_return())
        
        elif other == self.BOLD_ITALIC_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;3;9;", self.__prepare_return())
        
        elif other == self.BOLD_ITALIC_OVERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;3;53;", self.__prepare_return())
        
        elif other == self.BOLD_UNDERLINE_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;4;9;", self.__prepare_return())
        
        elif other == self.BOLD_UOLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[1;4;53;", self.__prepare_return())
        
        elif other == self.ITALIC_UNDERLINE_STRIKE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;4;9;", self.__prepare_return())
        
        elif other == self.ITALIC_UOLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;4;53;", self.__prepare_return())
        
        elif other == self.ITALIC_STRIKE_OVERLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[3;9;53;", self.__prepare_return())
        
        elif other == self.STRIKE_UOLINE:
            return _re.sub(r"^(\033\[|\u001b\[)", "\033[4;9;53;", self.__prepare_return())
        
        else:
            # replace error due to enumerator type change (0.3.27)
            if True:
                error = TypeError("expected one from following constant values as a right operand: " + repr(self.__constants__) + " or a specific lowercased string literal when single font style was meant to be applied")
            else:
                error = TypeError(
                    "Expected any from constant values: " + repr(self.__constants__) + ". You are discouraged to do common operations on these constants, like union as in case of regular expression flags, to satisfy this requirement, because it "
                    "won't warrant that returned string will be styled as thought"
                )
            raise error
    __dir__ = [n for n in locals() if n[:1] != "_"]
    "\\@since 0.3.26rc2"
    __all__ = [n for n in locals() if n[:1] != "_"]
    "\\@since 0.3.26rc2. Returns list of all non-underscore-preceded members of class `tense.Color`"
    __constants__ = [n for n in locals() if n[:1] != "_" and n.isupper()]
    
    """
    \\@since 0.3.26rc2

    Returns list of constants. These can be used as right operand for `%` operator. \\
    They are sorted as in ANSI escape code table, in ascending order
    """

if __name__ == "__main__":
    error = (RuntimeError, "this file is not for compiling, consider importing it instead")
    raise error

__all__ = sorted([n for n in globals() if n[:1] != "_"])
__dir__ = __all__

__author__ = "Aveyzan <aveyzan@gmail.com>"
"\\@since 0.3.26rc3"
__license__ = "MIT"
"\\@since 0.3.26rc3"
__version__ = Tense.version
"\\@since 0.3.26rc3"